# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2011-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy

from .. import config, math
from ..experiment import HXRD
from ..gridder import FuzzyGridder1D


def _get_cut(pos_along, pos_perp, intensity, dis, npoints):
    """
    obtain a line cut from 2D data using a FuzzyGridder1D to do the hard work.
    Data points with value of pos_perp smaller than `dis` will be considered in
    the line cut

    Parameters
    ----------
    pos_along :     array-like
        position along the cut which should be taken
    pos_perp :      array-like
        distance from the line cut axis. only data points with distance < `dis`
        will be considered
    intensity :     array-like
        data points, `pos_along`, `pos_perp`, and `intensity` must have the
        same shape
    dis :           float
        maximum distance to be allowed for contributing data points
    npoints :       int
        number of points in the output data

    Returns
    -------
    x :     ndarray
        gridded position along the cut axis
    d :     ndarray
        gridded data values for every position `x` along the cut line
    mask:   ndarray
        mask which is 1 for every used data point and 0 for the rest
    """
    pos_a = pos_along.ravel()
    pos_p = pos_perp.ravel()
    ma = numpy.abs(pos_p) < dis
    g1d = FuzzyGridder1D(npoints)
    width = (numpy.max(pos_a[ma]) - numpy.min(pos_a[ma])) / float(npoints)
    g1d(pos_a[ma], intensity.ravel()[ma], width=width)
    return g1d.xaxis, g1d.data, ma.astype(numpy.int8).reshape(intensity.shape)


def get_qz_scan(qpos, intensity, cutpos, npoints, intrange, **kwargs):
    r"""
    extracts a qz scan from reciprocal space map data with integration along
    either, the perpendicular plane in q-space, omega (sample rocking angle) or
    2theta direction. For the integration in angular space (omega, or 2theta)
    the coplanar diffraction geometry with qy and qz as diffraction plane is
    assumed. This is consistent with the coplanar geometry implemented in the
    HXRD-experiment class.

    This function works for 2D and 3D input data in the same way!

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of y, z (list with two components) or x, y, z (list with three
        components) momentum transfers
    intensity : array-like
        2D or 3D array of reciprocal space intensity with shape equal to the
        qpos entries
    cutpos :    float or tuple/list
        x/y-position at which the line scan should be extracted. this must be a
        float for 2D data and a tuple with two values for 3D data
    npoints :   int
        number of points in the output data
    intrange :  float
        integration range in along `intdir`, either in 1/\AA (`q`) or degree
        ('omega', or '2theta'). data will be integrated from
        `-intrange/2 .. +intrange/2`

    intdir :    {'q', 'omega', '2theta'}, optional
        integration direction: 'q': perpendicular Q-plane (default), 'omega':
        sample rocking angle, or '2theta': scattering angle.
    wl :       float or str, optional
        wavelength used to determine angular integration positions

    Note:
        For 3D data the angular integration directions although applicable for
        any set of data only makes sense when the data are aligned into the
        y/z-plane.

    Returns
    -------
    qz, qzint :     ndarray
        qz scan coordinates and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> qzcut, qzcut_int, mask = get_qz_scan([qy, qz], inten, 3.0, 200,
                                             intrange=0.3)
    """
    intdir = kwargs.get('intdir', 'q')
    lam = kwargs.get('wl', config.WAVELENGTH)
    hxrd = HXRD([1, 0, 0], [0, 0, 1], wl=lam)

    # make all data 3D
    if len(qpos) == 2:
        lqpos = [numpy.zeros_like(qpos[0]), qpos[0], qpos[1]]
        lcut = [0, cutpos]
    else:
        lqpos = qpos
        lcut = cutpos

    # make line cuts with selected integration direction
    if intdir == 'q':
        qperp = numpy.sqrt((lqpos[0]-lcut[0])**2 + (lqpos[1]-lcut[1])**2)
        ret = _get_cut(lqpos[2], qperp, intensity, intrange/2., npoints)
    elif intdir == 'omega':
        om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='realTilt')
        q = 4 * numpy.pi / lam * numpy.sin(numpy.radians(tt/2))
        ocut = numpy.degrees(numpy.arcsin(lcut[1]/q)) + tt / 2
        qzpos = q * numpy.cos(numpy.radians(ocut - tt/2))
        ret = _get_cut(qzpos, om-ocut, intensity, intrange/2., npoints)
    elif intdir == '2theta':
        om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='realTilt')
        q = 4 * numpy.pi / lam * numpy.sin(numpy.radians(tt/2))
        ttcut = 2 * (om - numpy.degrees(numpy.arcsin(lcut[1]/q)))
        qzpos = 4 * numpy.pi / lam * numpy.sin(numpy.radians(ttcut/2)) *\
            numpy.cos(numpy.radians(om - ttcut/2))
        ret = _get_cut(qzpos, tt-ttcut, intensity, intrange/2., npoints)

    return ret


def get_qy_scan(qpos, intensity, cutpos, npoints, intrange, **kwargs):
    r"""
    extracts a qy scan from reciprocal space map data with integration along
    either, the perpendicular plane in q-space, omega (sample rocking angle) or
    2theta direction. For the integration in angular space (omega, or 2theta)
    the coplanar diffraction geometry with qy and qz as diffraction plane is
    assumed. This is consistent with the coplanar geometry implemented in the
    HXRD-experiment class.

    This function works for 2D and 3D input data in the same way!

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of y, z (list with two components) or x, y, z (list with three
        components) momentum transfers
    intensity : array-like
        2D or 3D array of reciprocal space intensity with shape equal to the
        qpos entries
    cutpos :    float or tuple/list
        x/z-position at which the line scan should be extracted. this must be a
        float for 2D data (z-position) and a tuple with two values for 3D data
    npoints :   int
        number of points in the output data
    intrange :  float
        integration range in along `intdir`, either in 1/\AA (`q`) or degree
        ('omega', or '2theta'). data will be integrated from
        `-intrange .. +intrange`

    intdir :    {'q', 'omega', '2theta'}, optional
        integration direction: 'q': perpendicular Q-plane (default), 'omega':
        sample rocking angle, or '2theta': scattering angle.
    wl :       float or str, optional
        wavelength used to determine angular integration positions

    Note:
        For 3D data the angular integration directions although applicable for
        any set of data only makes sense when the data are aligned into the
        y/z-plane.

    Returns
    -------
    qy, qyint :     ndarray
        qy scan coordinates and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> qycut, qycut_int, mask = get_qy_scan([qy, qz], inten, 5.0, 250,
                                             intrange=0.02, intdir='2theta')
    """
    intdir = kwargs.get('intdir', 'q')
    lam = kwargs.get('wl', config.WAVELENGTH)
    hxrd = HXRD([1, 0, 0], [0, 0, 1], wl=lam)

    # make all data 3D
    if len(qpos) == 2:
        lqpos = [numpy.zeros_like(qpos[0]), qpos[0], qpos[1]]
        lcut = [0, cutpos]
    else:
        lqpos = qpos
        lcut = cutpos

    # make line cuts with selected integration direction
    if intdir == 'q':
        qperp = numpy.sqrt((lqpos[0]-lcut[0])**2 + (lqpos[2]-lcut[1])**2)
        ret = _get_cut(lqpos[1], qperp, intensity, intrange/2., npoints)
    elif intdir == 'omega':
        om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='real')
        q = 4 * numpy.pi / lam * numpy.sin(numpy.radians(tt/2))
        ocut = tt / 2 + (numpy.sign(lqpos[1].ravel()) *
                         numpy.degrees(numpy.arccos(lcut[1]/q)))
        qypos = q * numpy.sin(numpy.radians(ocut - tt/2))
        ret = _get_cut(qypos, om-ocut, intensity, intrange/2., npoints)
    elif intdir == '2theta':
        om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='real')
        ttcut = (om - numpy.degrees(numpy.arcsin(numpy.sin(numpy.radians(om)) -
                                    lcut[1]*lam/(2*numpy.pi)))) % 360
        ttcut2 = (om +
                  numpy.degrees(numpy.arcsin(numpy.sin(numpy.radians(om)) -
                                lcut[1]*lam/(2*numpy.pi))) +
                  180) % 360
        mask = numpy.abs(tt - om) > 90
        ttcut[mask] = ttcut2[mask]
        q = 4 * numpy.pi / lam * numpy.sin(numpy.radians(ttcut/2))
        qypos = q * numpy.sin(numpy.radians(om - ttcut/2))
        ret = _get_cut(qypos, tt-ttcut, intensity, intrange/2., npoints)

    return ret


def get_qx_scan(qpos, intensity, cutpos, npoints, intrange, **kwargs):
    r"""
    extracts a qx scan from 3D reciprocal space map data with integration along
    either, the perpendicular plane in q-space, omega (sample rocking angle) or
    2theta direction. For the integration in angular space (omega, or 2theta)
    the coplanar diffraction geometry with qy and qz as diffraction plane is
    assumed. This is consistent with the coplanar geometry implemented in the
    HXRD-experiment class.

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of x, y, z (list with three components) momentum transfers
    intensity : array-like
        3D array of reciprocal space intensity with shape equal to the
        qpos entries
    cutpos :    tuple/list
        y/z-position at which the line scan should be extracted. this must be
        and a tuple/list with the qy, qz cut position
    npoints :   int
        number of points in the output data
    intrange :  float
        integration range in along `intdir`, either in 1/\AA (`q`) or degree
        ('omega', or '2theta'). data will be integrated from
        `-intrange .. +intrange`

    intdir :    {'q', 'omega', '2theta'}, optional
        integration direction: 'q': perpendicular Q-plane (default), 'omega':
        sample rocking angle, or '2theta': scattering angle.
    wl :       float or str, optional
        wavelength used to determine angular integration positions

    Note:
        The angular integration directions although applicable for
        any set of data only makes sense when the data are aligned into the
        y/z-plane.

    Returns
    -------
    qx, qxint :     ndarray
        qx scan coordinates and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> qxcut, qxcut_int, mask = get_qx_scan([qx, qy, qz], inten, [0, 2.0],
                                             250, intrange=0.01)
    """
    intdir = kwargs.get('intdir', 'q')
    lam = kwargs.get('wl', config.WAVELENGTH)
    hxrd = HXRD([1, 0, 0], [0, 0, 1], wl=lam)

    # make line cuts with selected integration direction
    if intdir == 'q':
        qperp = numpy.sqrt((qpos[1]-cutpos[0])**2 + (qpos[2]-cutpos[1])**2)
        ret = _get_cut(qpos[0], qperp, intensity, intrange/2., npoints)
    elif intdir == 'omega':
        # needs testing
        om, chi, phi, tt = hxrd.Q2Ang(*qpos, trans=False, geometry='realTilt')
        ocut, dmy, dmy, ttcut = hxrd.Q2Ang(qpos[0], cutpos[0], cutpos[1],
                                           trans=False, geometry='realTilt')
        ret = _get_cut(qpos[0], om-ocut, intensity, intrange/2., npoints)
    elif intdir == '2theta':
        # needs testing
        om, chi, phi, tt = hxrd.Q2Ang(*qpos, trans=False, geometry='realTilt')
        ocut, dmy, dmy, ttcut = hxrd.Q2Ang(qpos[0], cutpos[0], cutpos[1],
                                           trans=False, geometry='realTilt')
        ret = _get_cut(qpos[0], tt-ttcut, intensity, intrange/2., npoints)

    return ret


def get_omega_scan(qpos, intensity, cutpos, npoints, intrange, **kwargs):
    """
    extracts an omega scan from reciprocal space map data with integration
    along either the 2theta, or radial (omega-2theta) direction. The coplanar
    diffraction geometry with qy and qz as diffraction plane is assumed. This
    is consistent with the coplanar geometry implemented in the HXRD-experiment
    class.

    This function works for 2D and 3D input data in the same way!

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of y, z (list with two components) or x, y, z (list with three
        components) momentum transfers
    intensity : array-like
        2D or 3D array of reciprocal space intensity with shape equal to the
        qpos entries
    cutpos :    tuple or list
        y/z-position or x/y/z-position at which the line scan should be
        extracted. this must be have two entries for 2D data (z-position) and a
        three for 3D data
    npoints :   int
        number of points in the output data
    intrange :  float
        integration range in along `intdir` in degree. data will be integrated
        from `-intrange .. +intrange`

    intdir :    {'2theta', 'radial'}, optional
        integration direction: '2theta': scattering angle (default), or
        'radial': omega-2theta direction.
    wl :       float or str, optional
        wavelength used to determine angular integration positions

    Note:
        Although applicable for any set of data, the extraction only makes
        sense when the data are aligned into the y/z-plane.

    Returns
    -------
    om, omint :     ndarray
        omega scan coordinates and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> omcut, omcut_int, mask = get_omega_scan([qy, qz], inten, [2.0, 5.0],
                                                250, intrange=0.1)
    """
    intdir = kwargs.get('intdir', '2theta')
    lam = kwargs.get('wl', config.WAVELENGTH)
    hxrd = HXRD([1, 0, 0], [0, 0, 1], wl=lam)

    # make all data 3D
    if len(qpos) == 2:
        lqpos = [numpy.zeros_like(qpos[0]), qpos[0], qpos[1]]
        lcut = [0, cutpos[0], cutpos[1]]
    else:
        lqpos = qpos
        lcut = cutpos

    # make line cuts with selected integration direction
    om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='realTilt')
    ocut, dmy, dmy, ttcut = hxrd.Q2Ang(*lcut, trans=False, geometry='realTilt')
    if intdir == '2theta':
        ret = _get_cut(om, tt-ttcut, intensity, intrange/2., npoints)
    elif intdir == 'radial':
        ret = _get_cut(om-(tt-ttcut)/2, tt-ttcut, intensity, intrange/2.,
                       npoints)

    return ret


def get_radial_scan(qpos, intensity, cutpos, npoints, intrange, **kwargs):
    """
    extracts a radial scan from reciprocal space map data with integration
    along either the omega or 2theta direction. The coplanar
    diffraction geometry with qy and qz as diffraction plane is assumed. This
    is consistent with the coplanar geometry implemented in the HXRD-experiment
    class.

    This function works for 2D and 3D input data in the same way!

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of y, z (list with two components) or x, y, z (list with three
        components) momentum transfers
    intensity : array-like
        2D or 3D array of reciprocal space intensity with shape equal to the
        qpos entries
    cutpos :    tuple or list
        y/z-position or x/y/z-position at which the line scan should be
        extracted. this must be have two entries for 2D data (z-position) and a
        three for 3D data
    npoints :   int
        number of points in the output data
    intrange :  float
        integration range in along `intdir` in degree. data will be integrated
        from `-intrange .. +intrange`

    intdir :    {'omega', '2theta'}, optional
        integration direction: 'omega': sample rocking angle (default),
        '2theta': scattering angle
    wl :       float or str, optional
        wavelength used to determine angular integration positions

    Note:
        Although applicable for any set of data, the extraction only makes
        sense when the data are aligned into the y/z-plane.

    Returns
    -------
    tt, omttint :     ndarray
        omega-2theta scan coordinates (2theta values) and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> ttcut, omtt_int, mask = get_radial_scan([qy, qz], inten, [2.0, 5.0],
                                                250, intrange=0.1)
    """
    intdir = kwargs.get('intdir', 'omega')
    lam = kwargs.get('wl', config.WAVELENGTH)
    hxrd = HXRD([1, 0, 0], [0, 0, 1], wl=lam)

    # make all data 3D
    if len(qpos) == 2:
        lqpos = [numpy.zeros_like(qpos[0]), qpos[0], qpos[1]]
        lcut = [0, cutpos[0], cutpos[1]]
    else:
        lqpos = qpos
        lcut = cutpos

    # make line cuts with selected integration direction
    om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='realTilt')
    ocut, dmy, dmy, ttcut = hxrd.Q2Ang(*lcut, trans=False, geometry='realTilt')
    if intdir == 'omega':
        ret = _get_cut(tt, om-((tt-ttcut)/2+ocut), intensity, intrange/2.,
                       npoints)
    elif intdir == '2theta':
        offcut = ttcut/2 - ocut
        ret = _get_cut(tt-2*(tt/2-om-offcut), 2*(tt/2-om-offcut), intensity,
                       intrange/2., npoints)

    return ret


def get_ttheta_scan(qpos, intensity, cutpos, npoints, intrange, **kwargs):
    """
    extracts a 2theta scan from reciprocal space map data with integration
    along either the omega or radial direction. The coplanar
    diffraction geometry with qy and qz as diffraction plane is assumed. This
    is consistent with the coplanar geometry implemented in the HXRD-experiment
    class.

    This function works for 2D and 3D input data in the same way!

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of y, z (list with two components) or x, y, z (list with three
        components) momentum transfers
    intensity : array-like
        2D or 3D array of reciprocal space intensity with shape equal to the
        qpos entries
    cutpos :    tuple or list
        y/z-position or x/y/z-position at which the line scan should be
        extracted. this must be have two entries for 2D data (z-position) and a
        three for 3D data
    npoints :   int
        number of points in the output data
    intrange :  float
        integration range in along `intdir` in degree. data will be integrated
        from `-intrange .. +intrange`

    intdir :    {'omega', 'radial'}, optional
        integration direction: 'omega': sample rocking angle (default),
        'radial': omega-2theta direction
    wl :       float or str, optional
        wavelength used to determine angular integration positions

    Note:
        Although applicable for any set of data, the extraction only makes
        sense when the data are aligned into the y/z-plane.

    Returns
    -------
    tt, ttint :     ndarray
        2theta scan coordinates and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> ttcut, tt_int, mask = get_ttheta_scan([qy, qz], inten, [2.0, 5.0],
                                              250, intrange=0.1)
    """
    intdir = kwargs.get('intdir', 'omega')
    lam = kwargs.get('wl', config.WAVELENGTH)
    hxrd = HXRD([1, 0, 0], [0, 0, 1], wl=lam)

    # make all data 3D
    if len(qpos) == 2:
        lqpos = [numpy.zeros_like(qpos[0]), qpos[0], qpos[1]]
        lcut = [0, cutpos[0], cutpos[1]]
    else:
        lqpos = qpos
        lcut = cutpos

    # make line cuts with selected integration direction
    om, chi, phi, tt = hxrd.Q2Ang(*lqpos, trans=False, geometry='realTilt')
    ocut, dmy, dmy, ttcut = hxrd.Q2Ang(*lcut, trans=False, geometry='realTilt')
    if intdir == 'omega':
        ret = _get_cut(tt, om-ocut, intensity, intrange/2., npoints)
    elif intdir == 'radial':
        ret = _get_cut(tt-2*(om-ocut), 2*(om-ocut), intensity,
                       intrange/2., npoints)

    return ret


def get_arbitrary_line(qpos, intensity, point, vec, npoints, intrange):
    """
    extracts a line scan from reciprocal space map data along an arbitrary line
    defined by the point 'point' and propergation vector 'vec'. Integration of
    the data is performed in a cylindrical volume along the line.
    This function works for 2D and 3D input data!

    Parameters
    ----------
    qpos :      list of array-like objects
        arrays of x, y (list with two components) or x, y, z (list with three
        components) momentum transfers
    intensity : array-like
        2D or 3D array of reciprocal space intensity with shape equal to the
        qpos entries
    point :     tuple, list or array-like
        point on the extraction line (2 or 3 coordinates)
    vec :       tuple, list or array-like
        propergation vector of the extraction line (2 or 3 coordinates)
    npoints :   int
        number of points in the output data
    intrange :  float
        radius of the cylindrical integration volume around the extraction line

    Returns
    -------
    qpos, qint :     ndarray
        line scan coordinates and intensities
    used_mask :     ndarray
        mask of used data, shape is the same as the input intensity: True for
        points which contributed, False for all others

    Examples
    --------
    >>> qcut, qint, mask = get_arbitrary_line([qx, qy, qz], inten,
                                              (1.1, 2.2, 0.0),
                                              (1, 1, 1), 200, 0.1)
    """
    # make all data 3D
    if len(qpos) == 2:
        lqpos = [numpy.zeros_like(qpos[0]), qpos[0], qpos[1]]
        lpoint = [0, point[0], point[1]]
        lvec = [0, vec[0], vec[1]]
    else:
        lqpos = qpos
        lpoint = point
        lvec = vec

    # make line cut
    lqpos = numpy.reshape(numpy.asarray([lqpos[0].ravel(), lqpos[1].ravel(),
                                         lqpos[2].ravel()]).T, (-1, 3))
    qalong = math.VecDot(lqpos, math.VecUnit(lvec))
    qdistance = math.distance(lqpos[:, 0], lqpos[:, 1], lqpos[:, 2], lpoint,
                              lvec)
    return _get_cut(qalong, qdistance, intensity, intrange, npoints)



print("import: 'xrayutilities'")
import xrayutilities

print("import: 'xrayutilities.analysis'")
import xrayutilities.analysis

print("import: 'xrayutilities.io'")
import xrayutilities.io

print("import: 'xrayutilities.materials'")
import xrayutilities.materials

print("import: 'xrayutilities.math'")
import xrayutilities.math

print("import: 'xrayutilities.simpack'")
import xrayutilities.simpack




# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2011-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
functions to help with experimental alignment during experiments, especially
for experiments with linear and area detectors
"""

import glob
import math
import numbers
import re
import time

import numpy
import scipy.optimize as optimize
from numpy import cos, degrees, radians, sin, tan
from scipy.ndimage.measurements import center_of_mass
from scipy.odr import models
from scipy.odr import odrpack as odr

from .. import config, cxrayutilities
from .. import math as xumath
from .. import utilities
from ..exception import InputError
from ..math import fwhm_exp

# regular expression to check goniometer circle syntax
circleSyntax = re.compile("[xyz][+-]")

#################################################
# channel per degree calculation
#################################################


def psd_chdeg(angles, channels, stdev=None, usetilt=True, plot=True,
              datap="xk", modelline="--r", modeltilt="-b", fignum=None,
              mlabel="fit", mtiltlabel="fit w/tilt", dlabel="data",
              figtitle=True):
    """
    function to determine the channels per degree using a linear
    fit of the function nchannel = center_ch+chdeg*tan(angles)
    or the equivalent including a detector tilt

    Parameters
    ----------
    angles :    array-like
        detector angles for which the position of the beam was measured
    channels :  array-like
        detector channels where the beam was found

    stdev :     array-like, optional
        standard deviation of the beam position
    plot :      bool, optional
        flag to specify if a visualization of the fit should be done
    usetilt :   bool, optional
        whether to use model considering a detector tilt, i.e. deviation angle
        of the pixel direction from orthogonal to the primary beam
        (default: True)

    Other Parameters
    ----------------
    datap :     str, optional
        plot format of data points
    modelline : str, optional
        plot format of modelline
    modeltilt : str, optional
        plot format of modeltilt
    fignum :    int or str, optional
        figure number to use for the plot
    mlabel :    str
        label of the model w/o tilt to be used in the plot
    mtiltlabel : str
        label of the model with tilt to be used in the plot
    dlabel :    str
        label of the data line to be used in the plot
    figtitle :  bool
        flag to tell if the figure title should show the fit parameters

    Returns
    -------
    pixelwidth :    float
        the width of one detector channel @ 1m distance, which is negative in
        case the hit channel number decreases upon an increase of the detector
        angle.
    centerch :      float
        center channel of the detector
    tilt :          float
        tilt of the detector from perpendicular to the beam (will be zero in
        case of usetilt=False)

    Note:
      L/pixelwidth*pi/180 = channel/degree for large detector distance with the
      sample detector disctance L
    """

    if stdev is None:
        stdevu = numpy.ones(len(channels))
    else:
        stdevu = stdev

    # define detector model and other functions needed for the tilt
    def straight_tilt(p, x):
        """
        model for straight-linear detectors including tilt

        Parameters
        ----------
        p :     list
            [L/w_pix*pi/180 ~= channel/degree, center_channel, detector_tilt]
            with L sample detector disctance, and w_pix the width of one
            detector channel
        x :     array-like
            independent variable of the model: detector angle (degree)
        """
        rad = radians(x)
        r = math.degrees(p[0]) * sin(rad) / \
            cos(rad - math.radians(p[2])) + p[1]
        return r

    def straight_tilt_der_x(p, x):
        """
        derivative of straight-linear detector model with respect to the angle
        for parameter description see straigt_tilt
        """
        rad = radians(x)
        p2 = math.radians(p[2])
        r = math.degrees(p[0]) * \
            (cos(rad) / cos(rad - p2) +
             sin(rad) / cos(rad - p2) ** 2 * sin(rad - p2))
        return r

    def straight_tilt_der_p(p, x):
        """
        derivative of straight-linear detector model with respect to the
        paramters for parameter description see straigt_tilt
        """
        rad = radians(x)
        p2 = math.radians(p[2])
        r = numpy.concatenate([degrees(sin(rad)) / cos(rad - p2),
                               numpy.ones(x.shape, dtype=float),
                               - math.degrees(p[0]) * sin(rad) /
                               cos(rad - p2) ** 2 * sin(rad - p2)])
        r.shape = (3,) + x.shape
        return r

    # fit linear
    model = models.unilinear
    data = odr.RealData(angles, channels, sy=stdevu)
    my_odr = odr.ODR(data, model)
    # fit type 2 for least squares
    my_odr.set_job(fit_type=2)
    fitlin = my_odr.run()

    # fit linear with tangens angle
    model = models.unilinear
    data = odr.RealData(degrees(tan(radians(angles))),
                        channels, sy=stdevu)
    my_odr = odr.ODR(data, model)
    # fit type 2 for least squares
    my_odr.set_job(fit_type=2)
    fittan = my_odr.run()

    if usetilt:
        # fit tilted straight detector model
        model = odr.Model(straight_tilt, fjacd=straight_tilt_der_x,
                          fjacb=straight_tilt_der_p)
        data = odr.RealData(angles, channels, sy=stdevu)
        my_odr = odr.ODR(data, model, beta0=[fittan.beta[0],
                                             fittan.beta[1], 0])
        # fit type 2 for least squares
        my_odr.set_job(fit_type=2)
        fittilt = my_odr.run()

    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis.psd_chdeg')

    if plot:
        markersize = 6.0
        markeredgewidth = 1.5
        linewidth = 2.0
        if fignum is None:
            plt.figure()
        else:
            plt.figure(fignum)
        # first plot to show linear model
        ax1 = plt.subplot(211)
        angr = angles.max() - angles.min()
        angp = numpy.linspace(angles.min() - angr * 0.1,
                              angles.max() + angr * .1, 1000)
        if modelline:
            plt.plot(angp, models._unilin(fittan.beta,
                                          degrees(tan(radians(angp)))),
                     modelline, label=mlabel, lw=linewidth)
        plt.plot(angp, models._unilin(fitlin.beta, angp), '-k', label='')
        if usetilt:
            plt.plot(angp, straight_tilt(fittilt.beta, angp),
                     modeltilt, label=mtiltlabel, lw=linewidth)
        if stdev is None:
            plt.plot(angles, channels, datap, ms=markersize,
                     mew=markeredgewidth, mec=datap[0],
                     mfc='none', label=dlabel)
        else:
            plt.errorbar(angles, channels, fmt=datap, yerr=stdevu,
                         ms=markersize, mew=markeredgewidth, mec=datap[0],
                         mfc='none', label=dlabel, ecolor='0.5')
        plt.grid(True)
        leg = plt.legend(numpoints=1)
        leg.get_frame().set_alpha(0.8)

        plt.ylabel("channel number")

        # lower plot to show deviations from linear model
        plt.subplot(212, sharex=ax1)
        if modelline:
            plt.plot(angp, models._unilin(fittan.beta,
                                          degrees(tan(radians(angp)))) -
                     models._unilin(fitlin.beta, angp),
                     modelline, label=mlabel, lw=linewidth)
        if usetilt:
            plt.plot(angp, straight_tilt(fittilt.beta, angp) -
                     models._unilin(fitlin.beta, angp),
                     modeltilt, label=mtiltlabel, lw=linewidth)
        if stdev is None:
            plt.plot(angles, channels - models._unilin(fitlin.beta, angles),
                     datap, ms=markersize, mew=markeredgewidth, mec=datap[0],
                     mfc='none', label=dlabel)
        else:
            plt.errorbar(angles,
                         channels - models._unilin(fitlin.beta, angles),
                         fmt=datap, yerr=stdevu, ms=markersize,
                         mew=markeredgewidth, mec=datap[0], mfc='none',
                         label=dlabel, ecolor='0.5')
        plt.xlabel("detector angle (deg)")
        plt.ylabel("ch. num. - linear trend")
        plt.grid(True)
        plt.hlines(0, angp.min(), angp.max())

        if figtitle:
            if usetilt:
                plt.suptitle(
                    "L/w*pi/180: %8.2f; center channel: %8.2f; tilt: %5.2fdeg"
                    % (fittilt.beta[0], fittilt.beta[1], fittilt.beta[2]))
            else:
                plt.suptitle("L/w*pi/180: %8.2f; center channel: %8.2f"
                             % (fittan.beta[0], fittan.beta[1]))

    if usetilt:
        fit = fittilt
    else:
        fit = fittan

    if config.VERBOSITY >= config.INFO_LOW:
        if usetilt:
            print("XU.analysis.psd_chdeg:  channelwidth@1m / center channel /"
                  " tilt: %8.4e / %8.2f / %6.3fdeg"
                  % (abs(1 / math.degrees(fit.beta[0])),
                     fit.beta[1], fit.beta[2]))
            print("XU.analysis.psd_chdeg:  error of channelwidth / "
                  "center channel / tilt: %8.4e / %8.3f / %6.3fdeg"
                  % (math.radians(fit.sd_beta[0] / fit.beta[0] ** 2),
                     fit.sd_beta[1], fit.sd_beta[2]))
        else:
            print("XU.analysis.psd_chdeg:  channelwidth@1m / center channel: "
                  "%8.4e / %8.2f"
                  % (1 / math.degrees(fit.beta[0]), fit.beta[1]))
            print("XU.analysis.psd_chdeg:  error of channelwidth / "
                  "center channel: %8.4e / %8.3f"
                  % (math.radians(fit.sd_beta[0] / fit.beta[0] ** 2),
                     fit.sd_beta[1]))

    if usetilt:
        return (1. / math.degrees(fit.beta[0]), fit.beta[1], fit.beta[2])
    else:
        return (1. / math.degrees(fit.beta[0]), fit.beta[1], 0.)


#################################################
# channel per degree calculation from scan with
# linear detector (determined maximum by peak_fit)
#################################################
def linear_detector_calib(angle, mca_spectra, **keyargs):
    """
    function to calibrate the detector distance/channel per degrees
    for a straight linear detector mounted on a detector arm

    Parameters
    ----------
    angle :     array-like
        array of angles in degree of measured detector spectra
    mca_spectra :   array-like
        corresponding detector spectra (shape: (len(angle), Nchannels)

    r_i :       str, optional
        primary beam direction as vector [xyz][+-]; default: 'y+'
    detaxis :   str, optional
        detector arm rotation axis [xyz][+-]; default: 'x+'

    Other parameters
    ----------------
    plot :      bool
        flag to specify if a visualization of the fit should be done
    usetilt :   bool
        whether to use model considering a detector tilt, i.e.  deviation angle
        of the pixel direction from orthogonal to the primary beam
        (default: True)

    Returns
    -------
    pixelwidth :    float
        width of the pixel at one meter distance, pixelwidth is negative in
        case the hit channel number decreases upon an increase of the detector
        angle
    center_channel : float
        central channel of the detector
    detector_tilt : float, optional
        if usetilt=True the fitted tilt of the detector is also returned

    Note:
      L/pixelwidth*pi/180 ~= channel/degree, with the sample detector
      distance L

    The function also prints out how a linear detector can be initialized using
    the results obtained from this calibration. Carefully check the results

    See Also
    --------
    psd_chdeg :  low level function with more configurable options
    """

    if "detaxis" in keyargs:
        detrotaxis = keyargs["detaxis"]
        keyargs.pop("detaxis")
    else:  # use default
        detrotaxis = 'x+'
    if "r_i" in keyargs:
        r_i = keyargs["r_i"]
        keyargs.pop("r_i")
    else:  # use default
        r_i = 'y+'

    # max intensity per spectrum
    mca_int = mca_spectra.sum(axis=1)
    mca_avg = numpy.average(mca_int)
    mca_rowmax = numpy.max(mca_int)
    mca_std = numpy.std(mca_int)

    # determine positions
    pos = []
    posstd = []
    ang = []
    nignored = 0
    for i in range(len(mca_spectra)):
        row = mca_spectra[i, :]
        row_int = row.sum()
        if ((abs(row_int - mca_avg) > 3 * mca_std) or
                (row_int - mca_rowmax * 0.7 < 0)):
            if config.VERBOSITY >= config.DEBUG:
                print("XU.analysis.linear_detector_calib: spectrum #%d "
                      "out of intensity range -> ignored" % i)
            nignored += 1
            continue

        maxp = numpy.argmax(row)
        fwhm = fwhm_exp(numpy.arange(row.size), row)
        N = int(7 * numpy.ceil(fwhm)) // 2 * 2

        # fit beam position
        # determine maximal usable length of array around peak position
        Nuse = min(maxp + N // 2, len(row) - 1) - max(maxp - N // 2, 0)
        param, perr, itlim = xumath.peak_fit(
            numpy.arange(Nuse),
            row[max(maxp - N // 2, 0):min(maxp + N // 2, len(row) - 1)],
            peaktype='PseudoVoigt')
        if param[0] > 0 and param[0] < Nuse and perr[0] < Nuse/2.:
            param[0] += max(maxp - N // 2, 0)
            pos.append(param[0])
            posstd.append(perr[0])
            ang.append(angle[i])

    ang = numpy.array(ang)
    pos = numpy.array(pos)
    posstd = numpy.array(posstd)
    if config.VERBOSITY >= config.INFO_ALL:
        print("XU.analysis.linear_detector_calib: using %d out of %d given "
              "spectra." % (len(ang), len(angle)))
    if config.VERBOSITY >= config.DEBUG:
        print("XU.analysis.linear_detector_calib: determined peak positions:")
        print(zip(pos, posstd))

    detparam = psd_chdeg(ang, pos, stdev=posstd, **keyargs)
    if numpy.sign(detparam[0]) > 0:
        sign = '-'
    else:
        sign = '+'

    detaxis = '  '
    detd = numpy.cross(xumath.getVector(detrotaxis), xumath.getVector(r_i))
    argm = numpy.abs(detd).argmax()

    def flipsign(char, val):
        if numpy.sign(val) < 0:
            if char == '+':
                return '-'
            else:
                return '+'
        else:
            return char

    if argm == 0:
        detaxis = 'x' + flipsign(sign, detd[argm])
    elif argm == 1:
        detaxis = 'y' + flipsign(sign, detd[argm])
    elif argm == 2:
        detaxis = 'z' + flipsign(sign, detd[argm])

    if config.VERBOSITY >= config.INFO_LOW:
        print("XU.analysis.linear_detector_calib:\n\tused/total spectra: %d/%d"
              % (mca_spectra.shape[0] - nignored, mca_spectra.shape[0]))
        print("\tdetector rotation axis (given by user/default input): %s"
              % detrotaxis)
        if len(detparam) == 3:
            tilt = detparam[2]
        else:
            tilt = 0
        print("\tdetector initialization with: init_linear('%s', %.2f, %d"
              ", pixelwidth=%.4e, distance=1., tilt=%.2f)"
              % (detaxis, abs(detparam[1]), mca_spectra.shape[1],
                 abs(detparam[0]), tilt))

    return detparam


######################################################
# detector parameter calculation from scan with
# area detector (determine maximum by center of mass)
######################################################
def area_detector_calib(angle1, angle2, ccdimages, detaxis, r_i, plot=True,
                        cut_off=0.7, start=(None, None, 1, 0, 0, 0, 0),
                        fix=(False, False, True, False, False, False, False),
                        fig=None, wl=None, plotlog=False, nwindow=50,
                        debug=False):
    """
    function to calibrate the detector parameters of an area detector
    it determines the detector tilt possible rotations and offsets in the
    detector arm angles

    Parameters
    ----------
    angle1 :    array-like
        outer detector arm angle
    angle2 :    array-like
        inner detector arm angle
    ccdimages : array-like
        images of the ccd taken at the angles given above
    detaxis :   list of str
        detector arm rotation axis; default: ['z+', 'y-']
    r_i :       str
        primary beam direction [xyz][+-]; default 'x+'

    plot :      bool, optional
        flag to determine if results and intermediate results should be
        plotted; default: True
    cut_off :   float, optional
        cut off intensity to decide if image is used for the determination or
        not; default: 0.7 = 70%
    start :     tuple, optional
        sequence of start values of the fit for parameters, which can not be
        estimated automatically or might want to be fixed.  These are: pwidth1,
        pwidth2, distance, tiltazimuth, tilt, detector_rotation,
        outerangle_offset.  By default (None, None, 1, 0, 0, 0, 0) is used.
    fix :       tuple of bool
        fix parameters of start (default: (False, False, True, False, False,
        False, False)) It is strongly recommended to either fix the distance or
        the pwidth1, 2 values.
    fig :       Figure, optional
        matplotlib figure used for plotting the error default: None (creates
        own figure)
    wl :        float or str
        wavelength of the experiment in angstrom (default: config.WAVELENGTH)
        value does not really matter here but does affect the scaling of the
        error
    plotlog :   bool
        flag to specify if the created error plot should be on log-scale
    nwindow :   int
        window size for determination of the center of mass position after the
        center of mass of every full image is determined, the center of mass is
        determined again using a window of size nwindow in order to reduce the
        effect of hot pixels.
    debug :     bool
        flag to specify that you want to see verbose output and saving of
        images to show if the CEN determination works
    """

    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis.area_'
                                                       'detector_calib')

    if wl is None:
        wl = config.WAVELENGTH
    else:
        wl = utilities.wavelength(wl)

    for i in [0, 1]:
        if fix[i] and not numpy.isscalar(start[i]):
            raise ValueError("XU.analysis.area_detector_calib: start value for"
                             " pwidth%d must be given if it should be fixed "
                             "during the fit" % (i+1))

    t0 = time.time()
    Npoints = len(angle1)
    if debug:
        print("number of given images: %d" % Npoints)

    # determine center of mass position from detector images
    # also use only images with an intensity larger than 70% of the average
    # intensity
    n1 = numpy.zeros(0, dtype=numpy.double)
    n2 = n1
    ang1 = n1
    ang2 = n1

    avg = 0
    for i in range(Npoints):
        avg += numpy.sum(ccdimages[i])
    avg /= float(Npoints)
    (N1, N2) = ccdimages[0].shape

    if debug:
        print("average intensity per image: %.1f" % avg)

    for i in range(Npoints):
        if debug and i == 0:
            print("angle1, angle2, cen1, cen2")
        img = ccdimages[i]
        if numpy.sum(img) > cut_off * avg:
            cen1, cen2 = _peak_position(img, nwindow, plot=debug and plot)
            n1 = numpy.append(n1, cen1)
            n2 = numpy.append(n2, cen2)
            ang1 = numpy.append(ang1, angle1[i])
            ang2 = numpy.append(ang2, angle2[i])
            if debug:
                print("%8.3f %8.3f \t%.2f %.2f" % (angle1[i], angle2[i],
                                                   cen1, cen2))
    Nused = len(ang1)

    if debug:
        print("Nused / Npoints: %d / %d" % (Nused, Npoints))

    # determine detector directions
    detdir1, detdir2 = _determine_detdir(
        ang1 - start[6], ang2, n1, n2, detaxis, r_i)

    if debug:
        print("determined detector directions:[%s, %s]" % (detdir1, detdir2))

    epslist = []
    paramlist = []
    epsmin = numpy.inf
    fitmin = None

    print("tiltaz   tilt   detrot   offset:  error (relative) (fittime)")
    print("------------------------------------------------------------")
    # find optimal detector rotation (however keep other parameters free)
    detrot = start[5]
    if not fix[5]:
        for detrotstart in numpy.linspace(start[5] - 1, start[5] + 1, 40):
            start = start[:5] + (detrotstart,) + (start[6],)
            eps, param, fit = _area_detector_calib_fit(
                ang1, ang2, n1, n2, detaxis, r_i, detdir1, detdir2,
                start=start, fix=fix, full_output=True, wl=wl, debug=debug)
            epslist.append(eps)
            paramlist.append(param)
            if epslist[-1] < epsmin:
                epsmin = epslist[-1]
                parammin = param
                fitmin = fit
                detrot = param[7]
            if debug:
                print(eps, param)

    Ntiltaz = 1 if fix[3] else 5
    Ntilt = 1 if fix[4] else 6
    Noffset = 1 if fix[6] else 100
    if fix[6]:
        Ntilt = Ntilt * 8 if not fix[4] else Ntilt
        Ntiltaz = Ntiltaz * 7 if not fix[3] else Ntiltaz

    startparam = start[:5] + (detrot,) + (start[6],)
    if debug:
        print("start params: %s" % str(startparam))

    Ntot = Ntiltaz * Ntilt * Noffset
    ict = 0
    for tiltazimuth in numpy.linspace(startparam[3] if fix[3] else 0,
                                      360, Ntiltaz, endpoint=False):
        for tilt in numpy.linspace(
                startparam[4] if fix[4] else max(0, startparam[4]-2),
                max(0, startparam[4]-2)+4, Ntilt):
            for offset in numpy.linspace(
                    startparam[6] if fix[6] else - 3 + startparam[6],
                    3 + startparam[6], Noffset):
                t1 = time.time()
                start = (startparam[0], startparam[1], startparam[2],
                         tiltazimuth, tilt, startparam[5], offset)
                eps, param, fit = _area_detector_calib_fit(
                    ang1, ang2, n1, n2, detaxis, r_i, detdir1, detdir2,
                    start=start, fix=fix, full_output=True, wl=wl)
                epslist.append(eps)
                paramlist.append(param)
                t2 = time.time()
                print("%d/%d\t%6.1f %6.2f %8.3f %8.3f: %10.4e (%4.2f) "
                      "(%5.2fsec)" % ((ict, Ntot) + start[3:] +
                                      (epslist[-1],
                                       epslist[-1] / epsmin, t2 - t1)))
                ict += 1

                if epslist[-1] < epsmin:
                    print("************************")
                    print("new global minimum found")
                    epsmin = epslist[-1]
                    parammin = param
                    fitmin = fit
                    print("new best parameters: %.2f %.2f %10.4e %10.4e %8.4f "
                          "%.1f %.2f %.3f %.3f" % parammin)
                    print("************************\n")

    (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt, detrot,
        outerangle_offset) = parammin

    if plot:
        if fig:
            plt.figure(fig.number)
        else:
            plt.figure("CCD Calib fit")
            plt.clf()
        nparams = numpy.array(paramlist)
        neps = numpy.array(epslist)
        labels = (
            'cch1 (1)',
            'cch2 (1)',
            r'pwidth1 ($\mu$m)',
            r'pwidth2 ($\mu$m)',
            'distance (m)',
            'tiltazimuth (deg)',
            'tilt (deg)',
            'detrot (deg)',
            'outerangle offset (deg)')
        xscale = (1., 1., 1.e6, 1.e6, 1., 1., 1., 1., 1.)
        for p in range(9):
            ax = plt.subplot(3, 3, p + 1)
            if plotlog:
                plt.semilogy(nparams[:, p] * xscale[p], neps, '.k')
            else:
                plt.scatter(nparams[:, p] * xscale[p], neps, c=nparams[:, -1],
                            s=10, marker='o', cmap=plt.cm.gnuplot,
                            edgecolor='none')
            plt.xlabel(labels[p])
            if plotlog:
                plt.semilogy(parammin[p] * xscale[p], epsmin, 'ok', ms=8,
                             mew=2.5, mec='k', mfc='w')
            else:
                plt.plot(parammin[p] * xscale[p], epsmin, 'ok', ms=8, mew=2.5,
                         mec='k', mfc='w')
                plt.ylim(epsmin * 0.7, epsmin * 2.)
            plt.locator_params(nbins=4, axis='x')
            if p > 1:
                if fix[p-2]:
                    ax.set_facecolor('0.85')
        plt.tight_layout()

    if config.VERBOSITY >= config.INFO_LOW:
        print("total time needed for fit: %.2fsec" % (time.time() - t0))
        print("fitted parameters: epsilon: %10.4e (%d,%s) "
              % (epsmin, fitmin.info, repr(fitmin.stopreason)))
        print("param: (cch1, cch2, pwidth1, pwidth2, tiltazimuth, tilt, "
              "detrot, outerangle_offset)")
        print("param: %.2f %.2f %10.4e %10.4e %.4f %.1f %.2f %.3f %.3f"
              % (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt,
                 detrot, outerangle_offset))

    if config.VERBOSITY > 0:
        print("please check the resulting data (consider setting plot=True)")
        print("detector rotation axis / primary beam direction "
              "(given by user): %s / %s" % (repr(detaxis), r_i))
        print("detector pixel directions / distance: %s %s / %g"
              % (detdir1, detdir2, 1.))
        print("\tdetector initialization with: init_area('%s', '%s', "
              "cch1=%.2f, cch2=%.2f, Nch1=%d, Nch2=%d, pwidth1=%.4e, "
              "pwidth2=%.4e, distance=%.5f, detrot=%.3f, tiltazimuth=%.1f, "
              "tilt=%.3f)" % (detdir1, detdir2, cch1, cch2, N1, N2, pwidth1,
                              pwidth2, distance, detrot, tiltazimuth, tilt))
        print("AND ALWAYS USE an (additional) OFFSET of %.4fdeg in the "
              "OUTER DETECTOR ANGLE!" % (outerangle_offset))

    return (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth,
            tilt, detrot, outerangle_offset), epsmin


def _peak_position(img, nwindow, plot=False):
    """
    function to determine the peak position on the detector using the center of
    mass (COM)

    Parameters
    ----------
    img :       array-like
        detector image data as 2D array
    nwindow :   int
        to avoid influence of hot pixels far away from the peak position the
        center of mass approach is repeated with a window around the COM of the
        full image. COM of the size (nwindow, nwindow) is returned
    plot :      bool, optional
        the result of the of the determination can be saved as a plot
    """
    nw = nwindow // 2
    [cen1r, cen2r] = center_of_mass(img)
    for i in range(11):  # refine center of mass multiple times
        [cen1, cen2] = center_of_mass(
            img[max(int(cen1r) - nw, 0):
                min(int(cen1r) + nw, img.shape[0]),
                max(int(cen2r) - nw, 0):
                min(int(cen2r) + nw, img.shape[1])])
        cen1 += max(int(cen1r) - nw, 0)
        cen2 += max(int(cen2r) - nw, 0)
        if numpy.linalg.norm((cen1 - cen1r, cen2 - cen2r)) > 3:
            cen1r, cen2r = (cen1, cen2)
        else:
            break
    if i == 10 and config.VERBOSITY >= config.INFO_LOW:
        print("XU.analysis._peak_position: Warning: peak position "
              "determination not converged, consider debug mode!")
    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis._peak_'
                                                       'position')
    if plot:
        plt.figure("_ccd")
        plt.imshow(utilities.maplog(img), origin='lower')
        plt.plot(cen2, cen1, 'ow', mfc='none')
        plt.axis([cen2 - nw, cen2 + nw, cen1 - nw, cen1 + nw])
        plt.colorbar()
        fnr = len(glob.glob('xu_calib_ccd_img*.png'))
        plt.savefig("xu_calib_ccd_img%d.png" % (fnr + 1))
        plt.close("_ccd")

    return cen1, cen2


def _determine_detdir(ang1, ang2, n1, n2, detaxis, r_i):
    """
    determines detector pixel direction from correlation analysis of linear
    fits to the observed pixel numbers of the primary beam.
    """
    # center channel and detector pixel direction and pixel size
    (s1, i1), r1 = xumath.linregress(ang1, n1)
    (s2, i2), r2 = xumath.linregress(ang1, n2)
    (s3, i3), r3 = xumath.linregress(ang2, n1)
    (s4, i4), r4 = xumath.linregress(ang2, n2)

    # determine detector directions
    s = ord('x') + ord('y') + ord('z')
    c1 = ord(detaxis[0][0]) + ord(r_i[0])
    c2 = ord(detaxis[1][0]) + ord(r_i[0])
    sign1 = numpy.sign(numpy.sum(numpy.cross(xumath.getVector(detaxis[0]),
                                             xumath.getVector(r_i))))
    sign2 = numpy.sign(numpy.sum(numpy.cross(xumath.getVector(detaxis[1]),
                                             xumath.getVector(r_i))))
    if ((r1 + r2 > r3 + r4) and r1 > r2) or ((r1 + r2 < r3 + r4) and r3 < r4):
        detdir1 = chr(s - c1)
        detdir2 = chr(s - c2)
        if numpy.sign(s1) > 0:
            if sign1 > 0:
                detdir1 += '-'
            else:
                detdir1 += '+'
        else:
            if sign1 > 0:
                detdir1 += '+'
            else:
                detdir1 += '-'

        if numpy.sign(s4) > 0:
            if sign2 > 0:
                detdir2 += '-'
            else:
                detdir2 += '+'
        else:
            if sign2 > 0:
                detdir2 += '+'
            else:
                detdir2 += '-'
    else:
        detdir1 = chr(s - c2)
        detdir2 = chr(s - c1)
        if numpy.sign(s3) > 0:
            if sign2 > 0:
                detdir1 += '-'
            else:
                detdir1 += '+'
        else:
            if sign2 > 0:
                detdir1 += '+'
            else:
                detdir1 += '-'

        if numpy.sign(s2) > 0:
            if sign1 > 0:
                detdir2 += '-'
            else:
                detdir2 += '+'
        else:
            if sign1 > 0:
                detdir2 += '+'
            else:
                detdir2 += '-'

    return detdir1, detdir2


def _area_detector_calib_fit(ang1, ang2, n1, n2, detaxis, r_i, detdir1,
                             detdir2, start=(None, None, 1, 0, 0, 0, 0),
                             fix=(False, False, True, False, False, False,
                                  False),
                             full_output=False, wl=1., debug=False):
    """
    INTERNAL FUNCTION
    function to calibrate the detector parameters of an area detector
    it determines the detector tilt possible rotations and offsets in the
    detector arm angles

    parameters
    ----------
    angle1 :    array-like
        outer detector arm angle
    angle2 :    array-like
        inner detector arm angle
    n1, n2 :    int
        pixel number at which the primary beam was observed
    detaxis :   list of str
        detector arm rotation axis; default: ['z+', 'y-']
    r_i :       str
        primary beam direction [xyz][+-]; default 'x+'
    detdir1, detdir2 :  str
        detector pixel directions of first and second pixel coordinates;
        e.g. 'y+'

    start :     tuple, optional
        sequence of start values of the fit for parameters, which can not be
        estimated automatically or might want to be fixed.  These are: pwidth1,
        pwidth2, distance, tiltazimuth, tilt, detector_rotation,
        outerangle_offset.  By default (None, None, 1, 0, 0, 0, 0) is used.
    fix :       tuple of bool
        fix parameters of start (default: (False, False, True, False, False,
        False, False)) It is strongly recommended to either fix the distance or
        the pwidth1, 2 values.
    fig :       Figure, optional
        matplotlib figure used for plotting the error default: None (creates
        own figure)
    full_output : bool
        flag to tell if to return fit object with final parameters and detector
        directions
    wl :        float or str
        wavelength of the experiment in angstrom (default: 1)
        value does not really matter here but does affect the scaling of the
        error
    debug :     bool
        flag to tell if you want to see debug output of the script (switch this
        to true only if you can handle it :))

    Returns
    -------
    float
        final epsilon of the fit
    param :     list, optional
        if full_output: fit parameters
    fit :       object, optional
        if full_output: fit object
    """

    def areapixel(params, detectorDir1, detectorDir2, r_i, detectorAxis,
                  *args, **kwargs):
        """
        angular to momentum space conversion for pixels of an area detector the
        center pixel is in direction of self.r_i when detector angles are zero

        the detector geometry must be given to the routine

        Parameters
        ----------
        params :    list
            parameters of the detector calibration model
            (cch1, cch2, pwidth1, pwidth2, tiltazimuth, tilt, detrot)

             - cch1, cch2: center pixel, in direction of self.r_i at zero
               detectorAngles;
             - pwidth1, pwidth2: width of one pixel (same unit as distance);
             - distance: distance of center pixel from center of rotation;
             - tiltazimuth: direction of the tilt vector in the detector plane
               (in degree);
             - tilt: tilt of the detector plane around an axis normal to the
               direction given by the tiltazimuth;
             - detrot: detector rotation around the primary beam direction as
               given by r_i;

        detectorDir1 :  str
            direction of the detector (along the pixel direction 1); e.g. 'z+'
            means higher pixel numbers at larger z positions
        detectorDir2 :  str
            direction of the detector (along the pixel direction 2); e.g. 'x+'
        r_i :           str
            primary beam direction e.g. 'x+'
        detectorAxis :  list or tuple
            detector circles e.g. ['z+', 'y-'] would mean a detector arm with a
            two rotations
        *args :     array-like
            detector angles and channel numbers;
            *dAngles* as numpy array, lists or Scalars in total
            len(detectorAxis) must be given starting with the outer most
            circle. All arguments must have the same shape or length.
            *channel numbers* n1 and n2 where the primary beam hits the
            detector with same length as the detector values

        delta :         list, optional
            giving delta angles to correct the given ones for misalignment
            delta must be an numpy array or list of len(*dAngles) used angles
            are than *args - delta
        wl :            float or str, optional
            x-ray wavelength in angstrom (default: 1 (since it does not matter
            here))
        deg :           bool, optional
            flag to tell if angles are passed as degree (default: True)

        Returns
        -------
        ndarray
            reciprocal space position of detector pixels n1, n2 in a
            numpy.ndarray of shape ( len(args) , 3 )
        """

        # check detector circle argument
        if isinstance(detectorAxis, (str, list, tuple)):
            if isinstance(detectorAxis, str):
                dAxis = list([detectorAxis])
            else:
                dAxis = list(detectorAxis)
            for circ in dAxis:
                if not isinstance(circ, str) or len(circ) != 2:
                    raise InputError("QConversionPixel: incorrect detector "
                                     "circle type or syntax (%s)" % repr(circ))
                if not circleSyntax.search(circ):
                    raise InputError("QConversionPixel: incorrect detector "
                                     "circle syntax (%s)" % circ)
        else:
            raise TypeError("Qconversion error: invalid type for detectorAxis,"
                            " must be str, list or tuple")
        # add detector rotation around primary beam
        dAxis += [r_i]
        _detectorAxis_str = ''
        for circ in dAxis:
            _detectorAxis_str += circ

        Nd = len(dAxis)
        Nargs = Nd + 2 - 1

        # check detectorDir
        if not isinstance(detectorDir1, str) or len(detectorDir1) != 2:
            raise InputError("QConversionPixel: incorrect detector direction1 "
                             "type or syntax (%s)" % repr(detectorDir1))
        if not circleSyntax.search(detectorDir1):
            raise InputError("QConversionPixel: incorrect detector direction1 "
                             "syntax (%s)" % detectorDir1)
        _area_detdir1 = detectorDir1
        if not isinstance(detectorDir2, str) or len(detectorDir2) != 2:
            raise InputError("QConversionPixel: incorrect detector direction2 "
                             "type or syntax (%s)" % repr(detectorDir2))
        if not circleSyntax.search(detectorDir2):
            raise InputError("QConversionPixel: incorrect detector direction2 "
                             "syntax (%s)" % detectorDir2)
        _area_detdir2 = detectorDir2

        # parse parameter arguments
        _area_cch1 = float(params[0])
        _area_cch2 = float(params[1])
        _area_pwidth1 = float(params[2])
        _area_pwidth2 = float(params[3])
        _area_distance = float(params[4])
        _area_tiltazimuth = math.radians(params[5])
        _area_tilt = math.radians(params[6])
        _area_rot = float(params[7])
        _area_ri = xumath.getVector(r_i) * _area_distance

        # kwargs
        wl = utilities.wavelength(kwargs.get('wl', 1.))
        deg = kwargs.get('deg', True)

        delta = numpy.asarray(kwargs.get('delta', numpy.zeros(Nd)),
                              dtype=numpy.double)
        if delta.size != Nd - 1:
            raise InputError("QConversionPixel: keyword argument delta "
                             "does not have an appropriate shape")

        # prepare angular arrays from *args
        # need one sample angle and one detector angle array
        if len(args) != Nargs:
            raise InputError("QConversionPixel: wrong amount (%d) of arguments"
                             " given, number of arguments should be %d"
                             % (len(args), Nargs))

        try:
            Npoints = len(args[0])
        except (TypeError, IndexError):
            Npoints = 1

        dAngles = numpy.array((), dtype=numpy.double)
        for i in range(Nd - 1):
            arg = args[i]
            if not isinstance(arg, (numbers.Number, tuple,
                                    list, numpy.ndarray)):
                raise TypeError("QConversionPixel: invalid type for one of "
                                "the detector coordinates, must be scalar, "
                                "list or array")
            elif isinstance(arg, numbers.Number):
                arg = numpy.array([arg], dtype=numpy.double)
            elif isinstance(arg, list):
                arg = numpy.array(arg, dtype=numpy.double)
            arg = arg - delta[i]
            dAngles = numpy.concatenate((dAngles, arg))
        # add detector rotation around primary beam
        dAngles = numpy.concatenate((dAngles,
                                     numpy.ones(arg.shape,
                                                dtype=numpy.double) *
                                     _area_rot))

        # read channel numbers
        n1 = numpy.array((), dtype=numpy.double)
        n2 = numpy.array((), dtype=numpy.double)

        arg = args[Nd - 1]
        if not isinstance(arg, (numbers.Number, tuple, list, numpy.ndarray)):
            raise TypeError("QConversionPixel: invalid type for one of the "
                            "detector coordinates, must be scalar, list or "
                            "array")
        elif isinstance(arg, numbers.Number):
            arg = numpy.array([arg], dtype=numpy.double)
        elif isinstance(arg, list):
            arg = numpy.array(arg, dtype=numpy.double)
        n1 = arg

        arg = args[Nd]
        if not isinstance(arg, (numbers.Number, tuple, list, numpy.ndarray)):
            raise TypeError("QConversionPixel: invalid type for one of the "
                            "detector coordinates, must be scalar, list or "
                            "array")
        elif isinstance(arg, numbers.Number):
            arg = numpy.array([arg], dtype=numpy.double)
        elif isinstance(arg, list):
            arg = numpy.array(arg, dtype=numpy.double)
        n2 = arg

        dAngles.shape = (Nd, Npoints)
        dAngles = dAngles.transpose()

        if deg:
            dAngles = radians(dAngles)

        qpos = cxrayutilities.ang2q_conversion_area_pixel(
            dAngles, n1, n2, _area_ri, _detectorAxis_str,
            _area_cch1, _area_cch2, _area_pwidth1, _area_pwidth2,
            _area_detdir1, _area_detdir2, _area_tiltazimuth, _area_tilt,
            wl, config.NTHREADS)

        return qpos[:, 0], qpos[:, 1], qpos[:, 2]

    def afunc(param, x, detectorDir1, detectorDir2, r_i, detectorAxis, wl):
        """
        function for fitting the detector parameters
        basically this is a wrapper for the areapixel function

        parameters
        ----------
        param :         list
            fit parameters (cch1, cch2, pwidth1, pwidth2, distance,
            tiltazimuth, tilt, detrot, outerangle_offset)
        x :             array-like
            independent variables (angle1, angle2, n1, n2) with
            shape (4, Npoints)
        detectorDir1 :  str
            direction of the detector (along the pixel direction 1); e.g. 'z+'
            means higher pixel numbers at larger z positions
        detectorDir2 :  str
            direction of the detector (along the pixel direction 2); e.g. 'x+'
        r_i :           str
            primary beam direction e.g. 'x+'
        detectorAxis :  list or tuple
            detector circles e.g. ['z+', 'y-'] would mean a detector arm with a
            two rotations
        wl :            float or str
            wavelength of the experiment in angstrom

        Returns
        -------
        ndarray
            reciprocal space position of detector pixels n1, n2 in a
            numpy.ndarray of shape (3, x.shape[1])
        """

        angle1 = x[0, :]
        angle2 = x[1, :]
        n1 = x[2, :]
        n2 = x[3, :]

        # use only positive tilt
        param[6] = abs(param[6])

        (qx, qy, qz) = areapixel(
            param[:-1], detectorDir1, detectorDir2, r_i, detectorAxis,
            angle1, angle2, n1, n2, delta=[param[-1], 0.], wl=wl)

        return qx ** 2 + qy ** 2 + qz ** 2

    Npoints = len(ang1)

    # guess initial parameters
    # center channel and detector pixel direction and pixel size
    (s1, i1), r1 = xumath.linregress(ang1 - start[6], n1)
    (s2, i2), r2 = xumath.linregress(ang1 - start[6], n2)
    (s3, i3), r3 = xumath.linregress(ang2, n1)
    (s4, i4), r4 = xumath.linregress(ang2, n2)

    distance = start[2]
    if ((r1 + r2 > r3 + r4) and r1 > r2) or ((r1 + r2 < r3 + r4) and r3 < r4):
        cch1 = i1
        cch2 = i4
        pwidth1 = 2 * distance / numpy.abs(s1) * math.tan(math.radians(0.5))
        pwidth2 = 2 * distance / numpy.abs(s4) * math.tan(math.radians(0.5))
    else:
        cch1 = i3
        cch2 = i2
        pwidth1 = 2 * distance / numpy.abs(s3) * math.tan(math.radians(0.5))
        pwidth2 = 2 * distance / numpy.abs(s2) * math.tan(math.radians(0.5))
    if numpy.isscalar(start[0]):
        pwidth1 = start[0]
    if numpy.isscalar(start[1]):
        pwidth2 = start[1]
    if numpy.isscalar(start[0]) or numpy.isscalar(start[1]):
        # find biggest correlation and recalculate distance
        idxmax = numpy.argmax((r1, r2, r3, r4))
        s = (s1, s2, s3, s4)[idxmax]
        distance = abs(s) / math.tan(math.radians(0.5)) / 2
        distance *= pwidth1 if idxmax < 2 else pwidth2
    tilt = abs(start[4])
    tiltazimuth = start[3]
    detrot = start[5]
    outerangle_offset = start[6]
    # parameters for the fitting
    param = (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt,
             detrot, outerangle_offset)
    if debug:
        print("initial parameters: ")
        print("primary beam / detector pixel directions / distance: "
              "%s / %s %s / %e" % (r_i, detdir1, detdir2, distance))
        print("param: (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, "
              "tilt, detrot, outerangle_offset)")
        print("param: %.2f %.2f %10.4e %10.4e %.3f %.1f %.2f %.3f %.3f"
              % param)

    # set data
    x = numpy.empty((4, Npoints), dtype=numpy.double)
    x[0, :] = ang1
    x[1, :] = ang2
    x[2, :] = n1
    x[3, :] = n2
    data = odr.Data(x, y=1)
    # define model for fitting
    model = odr.Model(afunc, extra_args=(detdir1, detdir2, r_i, detaxis, wl),
                      implicit=True)
    # check if parameters need to be fixed
    ifixb = ()
    for i in range(len(fix)):
        ifixb += (int(not fix[i]),)

    my_odr = odr.ODR(data, model, beta0=param, ifixb=(1, 1) + ifixb,
                     ifixx=(0, 0, 0, 0),
                     stpb=(0.4, 0.4, pwidth1/50., pwidth2/50.,
                           distance/1000, 2, 0.125, 0.01, 0.01),
                     sclb=(1/abs(cch1), 1/abs(cch2),
                           1/pwidth1, 1/pwidth2, 1/distance, 1/90.,
                           1/0.2, 1/0.2, 1/0.2),
                     maxit=200, ndigit=12, sstol=1e-11, partol=1e-11)
    if debug:
        my_odr.set_iprint(final=1)
        my_odr.set_iprint(iter=2)

    fit = my_odr.run()

    (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt, detrot,
        outerangle_offset) = fit.beta
    # fix things in parameters
    tiltazimuth = tiltazimuth % 360.
    tilt = abs(tilt)

    final_q = afunc([cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt,
                     detrot, outerangle_offset],
                    x, detdir1, detdir2, r_i, detaxis, wl)
    final_error = numpy.mean(final_q)

    if debug:
        print("fitted parameters: (%e, %d, %s) " % (final_error, fit.info,
                                                    repr(fit.stopreason)))
        print("primary beam / detector pixel directions / distance: "
              "%s / %s %s / %e" % (r_i, detdir1, detdir2, distance))
        print("param: (cch1, cch2, pwidth1, pwidth2, disance, tiltazimuth, "
              "tilt, detrot, outerangle_offset)")
        print("param: %.2f %.2f %10.4e %10.4e %.3f %.1f %.2f %.3f %.3f"
              % (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt,
                 detrot, outerangle_offset))

    if full_output:
        return final_error, (cch1, cch2, pwidth1, pwidth2, distance,
                             tiltazimuth, tilt, detrot, outerangle_offset), fit
    else:
        return final_error


# #####################################################
#  detector parameter calculation from scan with
#  area detector (determine maximum by center of mass)
# #####################################################
def area_detector_calib_hkl(sampleang, angle1, angle2, ccdimages, hkls,
                            experiment, material, detaxis, r_i, plot=True,
                            cut_off=0.7,
                            start=(None, None, 1, 0, 0, 0, 0, 0, 0, 'config'),
                            fix=(False, False, True, False, False, False,
                                 False, False, False, False),
                            fig=None, plotlog=False, nwindow=50, debug=False):
    """
    function to calibrate the detector parameters of an area detector
    it determines the detector tilt possible rotations and offsets in the
    detector arm angles

    in this variant not only scans through the primary beam but also scans at a
    set of symmetric reflections can be used for the detector parameter
    determination. for this not only the detector parameters but in addition
    the sample orientation and wavelength need to be fit.  Both images from the
    primary beam hkl = (0, 0, 0) and from a symmetric reflection
    hkl = (h, k, l) need to be given for a successful run.

    Parameters
    ----------
    sampleang : array-like
        sample rocking angle (needed to align the reflections (same rotation
        direction as inner detector rotation)) other sample angle are not
        allowed to be changed during the scans
    angle1 :    array-like
        outer detector arm angle
    angle2 :    array-like
        inner detector arm angle
    ccdimages : array-like
        images of the ccd taken at the angles given above
    hkls :      list or array-like
        hkl values for every image
    experiment : Experiment
        Experiment class object needed to get the UB matrix for the hkl peak
        treatment
    material :  Crystal
        material used as reference crystal
    detaxis :   list of str
        detector arm rotation axis; default: ['z+', 'y-']
    r_i :       str
        primary beam direction [xyz][+-]; default 'x+'

    plot :      bool, optional
        flag to determine if results and intermediate results should be
        plotted; default: True
    cut_off :   float, optional
        cut off intensity to decide if image is used for the determination or
        not; default: 0.7 = 70%
    start :     tuple, optional
        sequence of start values of the fit for parameters, which can not be
        estimated automatically or might want to be fixed.  These are: pwidth1,
        pwidth2, distance, tiltazimuth, tilt, detector_rotation,
        outerangle_offset, sampletilt, sampletiltazimuth, wavelength.  By
        default (None, None, 1, 0, 0, 0, 0, 0, 0, 'config').
    fix :       tuple of bool
        fix parameters of start (default: (False, False, True, False, False,
        False, False, False, False, False)) It is strongly recommended to
        either fix the distance or the pwidth1, 2 values.
    fig :       Figure, optional
        matplotlib figure used for plotting the error default: None (creates
        own figure)
    plotlog :   bool
        flag to specify if the created error plot should be on log-scale
    nwindow :   int
        window size for determination of the center of mass position after the
        center of mass of every full image is determined, the center of mass is
        determined again using a window of size nwindow in order to reduce the
        effect of hot pixels.
    debug :     bool
        flag to specify that you want to see verbose output and saving of
        images to show if the CEN determination works
    """
    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis.area_'
                                                       'detector_calib_hkl')

    if start[-1] == 'config':
        start[-1] = config.WAVELENGTH
    elif isinstance(start[-1], str):
        start[-1] = utilities.wavelength(start[-1])

    t0 = time.time()
    Npoints = len(angle1)
    if debug:
        print("number of given images: %d" % Npoints)

    # determine center of mass position from detector images also use only
    # images with an intensity larger than cut_off of the average intensity.
    # the image selection is only performed for images in the primary beam
    n1 = numpy.zeros(0, dtype=numpy.double)
    n2 = n1
    ang1 = n1
    ang2 = n1
    sang = n1
    usedhkls = []

    avg = 0
    imgpbcnt = 0
    for i in range(Npoints):
        if (numpy.all(hkls[i] == (0, 0, 0))):
            avg += numpy.sum(ccdimages[i])
            imgpbcnt += 1

    if imgpbcnt > 0:
        avg /= float(imgpbcnt)
    else:
        raise ValueError("XU.analyis.area_detector_calib_hkl: no required "
                         "images in the primary beam given!")
    (N1, N2) = ccdimages[0].shape

    if debug:
        print("average intensity per image in the primary beam: %.1f" % avg)

    for i in range(Npoints):
        if debug and i == 0:
            print("angle1, angle2, cen1, cen2")
        img = ccdimages[i]
        if ((numpy.sum(img) > cut_off * avg) or
                (numpy.all(hkls[i] != (0, 0, 0)))):
            cen1, cen2 = _peak_position(img, nwindow, plot=debug and plot)

            n1 = numpy.append(n1, cen1)
            n2 = numpy.append(n2, cen2)
            ang1 = numpy.append(ang1, angle1[i])
            ang2 = numpy.append(ang2, angle2[i])
            sang = numpy.append(sang, sampleang[i])
            usedhkls.append(hkls[i])
            if debug:
                print("%8.3f %8.3f \t%.2f %.2f" % (angle1[i], angle2[i],
                                                   cen1, cen2))

    Nused = len(ang1)
    usedhkls = numpy.array(usedhkls)

    if debug:
        print("Nused / Npoints: %d / %d" % (Nused, Npoints))

    # guess initial parameters
    n10 = numpy.zeros(0, dtype=numpy.double)
    n20 = n10
    ang10 = n10
    ang20 = n10
    for i in range(Nused):
        if numpy.all(usedhkls[i] == (0, 0, 0)):
            n10 = numpy.append(n10, n1[i])
            n20 = numpy.append(n20, n2[i])
            ang10 = numpy.append(ang10, ang1[i])
            ang20 = numpy.append(ang20, ang2[i])

    detdir1, detdir2 = _determine_detdir(ang10 - start[3], ang20, n10, n20,
                                         detaxis, r_i)

    epslist = []
    paramlist = []
    epsmin = numpy.inf
    fitmin = None

    print("tiltaz   tilt   detrot   offset  sampletilt+azimuth wavelength:  "
          "error (relative) (fittime)")
    print("------------------------------------------------------------")
    # find optimal detector rotation (however keep other parameters free)
    detrot = start[5]
    if not fix[5]:
        for detrotstart in numpy.linspace(start[5] - 1, start[5] + 1, 20):
            start = start[:5] + (detrotstart,) + start[6:]
            eps, param, fit = _area_detector_calib_fit2(
                sang, ang1, ang2, n1, n2, usedhkls, experiment, material,
                detaxis, r_i, detdir1, detdir2, start=start, fix=fix,
                full_output=True)
            epslist.append(eps)
            paramlist.append(param)
            if epslist[-1] < epsmin:
                epsmin = epslist[-1]
                parammin = param
                fitmin = fit
                detrot = param[7]
            if debug:
                print("single fit")
                print(param)

    Ntiltaz = 1 if fix[3] else 5
    Ntilt = 1 if fix[4] else 6
    Noffset = 1 if fix[6] else 100
    if fix[6]:
        Ntilt = Ntilt * 8 if not fix[4] else Ntilt
        Ntiltaz = Ntiltaz * 7 if not fix[3] else Ntiltaz

    startparam = start[:5] + (detrot,) + start[6:]

    Ntot = Ntiltaz * Ntilt * Noffset
    ict = 0
    for tiltazimuth in numpy.linspace(startparam[3] if fix[3] else 0, 360,
                                      Ntiltaz, endpoint=False):
        for tilt in numpy.linspace(startparam[4] if fix[4] else 0, 4, Ntilt):
            for offset in numpy.linspace(
                    startparam[6] if fix[6] else - 3 + startparam[6],
                    3 + startparam[6], Noffset):
                t1 = time.time()
                start = (start[:3] + (tiltazimuth, tilt, detrot, offset) +
                         start[7:])
                eps, param, fit = _area_detector_calib_fit2(
                    sang, ang1, ang2, n1, n2, usedhkls, experiment, material,
                    detaxis, r_i, detdir1, detdir2, start=start, fix=fix,
                    full_output=True)
                epslist.append(eps)
                paramlist.append(param)
                t2 = time.time()
                print("%d/%d\t%6.1f %6.2f %8.3f %8.3f %8.3f %7.2f %8.4f: "
                      "%10.4e (%4.2f) (%5.2fsec)"
                      % ((ict, Ntot) + start[3:] +
                         (epslist[-1], epslist[-1] / epsmin, t2 - t1)))
                ict += 1

                if epslist[-1] < epsmin:
                    print("************************")
                    print("new global minimum found")
                    epsmin = epslist[-1]
                    parammin = param
                    fitmin = fit
                    print("new best parameters: %.2f %.2f %10.4e %10.4e %8.4f "
                          "%.1f %.2f %.3f %.3f %.3f %.2f %.4f" % parammin)
                    print("************************\n")

    (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt, detrot,
     outerangle_offset, stilt, stazimuth, wavelength) = parammin

    if plot:
        if fig:
            plt.figure(fig.number)
        else:
            figlabel = "CCD Calib fit %d"
            i = 1
            while figlabel % i in plt.get_figlabels():
                i += 1
            plt.figure(figlabel % i)
        nparams = numpy.array(paramlist)
        neps = numpy.array(epslist)
        labels = (
            'cch1 (1)',
            'cch2 (1)',
            r'pwidth1 ($\mu$m@1m)',
            r'pwidth2 ($\mu$m@1m)',
            'distance (m)',
            'tiltazimuth (deg)',
            'tilt (deg)',
            'detrot (deg)',
            'outerangle offset (deg)',
            'sample tilt (deg)',
            'st azimuth (deg)',
            'wavelength (AA)')
        xscale = (1., 1., 1.e6, 1.e6, 1., 1., 1., 1., 1., 1., 1., 1.)
        for p in range(12):
            ax = plt.subplot(3, 4, p + 1)
            if plotlog:
                plt.semilogy(nparams[:, p] * xscale[p], neps, '.k')
            else:
                plt.scatter(nparams[:, p] * xscale[p], neps, c=nparams[:, -1],
                            s=10, marker='o', cmap=plt.cm.gnuplot,
                            edgecolor='none')
            plt.xlabel(labels[p])
            if plotlog:
                plt.semilogy(parammin[p] * xscale[p], epsmin, 'ok',
                             ms=8, mew=2.5, mec='k', mfc='w')
            else:
                plt.plot(parammin[p] * xscale[p], epsmin, 'ok',
                         ms=8, mew=2.5, mec='k', mfc='w')
                plt.ylim(epsmin * 0.7, epsmin * 2.)
            plt.locator_params(nbins=4, axis='x')
            if p > 1:
                if fix[p-2]:
                    ax.set_facecolor('0.85')
        plt.tight_layout()

    if config.VERBOSITY >= config.INFO_LOW:
        print("total time needed for fit: %.2fsec" % (time.time() - t0))
        print("fitted parameters: epsilon: %10.4e (%d,%s) "
              % (epsmin, fitmin.info, repr(fitmin.stopreason)))
        print("param: (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, "
              "tilt, detrot, outerangle_offset, sampletilt, stazimuth, "
              "wavelength)")
        print("param: %.2f %.2f %10.4e %10.4e %.4f %.1f %.2f %.3f %.3f %.3f "
              "%.2f %.4f" % (cch1, cch2, pwidth1, pwidth2, distance,
                             tiltazimuth, tilt, detrot, outerangle_offset,
                             stilt, stazimuth, wavelength))

    if config.VERBOSITY > 0:
        print("please check the resulting data (consider setting plot=True)")
        print("detector rotation axis / primary beam direction (given by user)"
              ": %s / %s" % (repr(detaxis), r_i))
        print("detector pixel directions / distance: %s %s / %e"
              % (detdir1, detdir2, distance))
        print("\tdetector initialization with: init_area('%s', '%s', "
              "cch1=%.2f, cch2=%.2f, Nch1=%d, Nch2=%d, pwidth1=%.4e, "
              "pwidth2=%.4e, distance=%.5f, detrot=%.3f, tiltazimuth=%.1f, "
              "tilt=%.3f)" % (detdir1, detdir2, cch1, cch2, N1, N2, pwidth1,
                              pwidth2, distance, detrot, tiltazimuth, tilt))
        print("AND ALWAYS USE an (additional) OFFSET of %.4fdeg in the "
              "OUTER DETECTOR ANGLE!" % (outerangle_offset))

    return (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth,
            tilt, detrot, outerangle_offset, stilt, stazimuth,
            wavelength), epsmin


def _area_detector_calib_fit2(sang, ang1, ang2, n1, n2, hkls, experiment,
                              material, detaxis, r_i, detdir1, detdir2,
                              start=(None, None, 1, 0, 0, 0, 0, 0, 0, 1.0),
                              fix=(False, False, True, False, False, False,
                                   False, False, False, False),
                              full_output=False, debug=False):
    """
    INTERNAL FUNCTION
    function to calibrate the detector parameters of an area detector
    it determines the detector tilt possible rotations and offsets in the
    detector arm angles

    Parameters
    ----------
    sang :      array-like
        sample rocking angle (rotation direction of inner detector rotation)
    angle1 :    array-like
        outer detector arm angle
    angle2 :    array-like
        inner detector arm angle
    n1, n2 :    array-like
        pixel number at which the beam was observed
    hkls :      tuple or list
        Miller indices of the reflection were the images were taken (use
        (0, 0, 0)) for primary beam
    experiment : Experiment
        Experiment class object needed to get the UB matrix needed for the hkl
        peak treatment
    material :  Crystal
        material used as reference crystal
    detaxis :   str
        detector arm rotation axis default: ['z+', 'y-']
    detdir1, detdir2 : str
        detector pixel directions of first and second pixel coordinates;
        e.g. 'y+'
    r_i :       str
        primary beam direction [xyz][+-] default 'x+'

    start :     tuple or list, optional
        sequence of start values of the fit for parameters, which can not be
        estimated automatically.  these are: pwidth1, pwidth2, distance,
        tiltazimuth, tilt, detector_rotation, outerangle_offset, sampletilt,
        sampletiltazimuth, wavelength.
        By default: (None, None, 1, 0, 0, 0, 0, 0, 0, 1.0)
    fix :       tuple or list, optional
        fix parameters of start
    full_output : bool, optional
        flag to tell if to return fit object with final parameters and detector
        directions
    debug :     bool, optional
        flag to tell if you want to see debug output of the script (switch this
        to true only if you can handle it :))

    Returns
    -------
    float
        final epsilon of the fit
    param :     list, optional
        if full_output: fit parameters
    fit :       object, optional
        if full_output: fit object
    """

    def areapixel2(params, detectorDir1, detectorDir2, r_i, detectorAxis,
                   *args, **kwargs):
        """
        angular to momentum space conversion for pixels of an area detector the
        center pixel is in direction of self.r_i when detector angles are zero

        the detector geometry must be given to the routine

        Parameters
        ----------
        params :        list or tuple
            parameters of the detector calibration model
            (cch1, cch2, pwidth1, pwidth2, tiltazimuth, tilt, detrot):
            cch1, 2: center pixel, in direction of r_i at zero detectorAngles;
            pwidth1, 2: width of one pixel (same unit as distance);
            distance: distance of center pixel from center of rotation;
            tiltazimuth: direction of the tilt vector in the detector plane (in
            degree);
            tilt: tilt of the detector plane around an axis normal to the
            direction given by the tiltazimuth;
            detrot: detector rotation around the primary beam direction as
            given by r_i
        detectorDir1 :  str
            direction of the detector (along the pixel direction 1); e.g. 'z+'
            means higher pixel numbers at larger z positions
        detectorDir2 :  str
            direction of the detector (along the pixel direction 2); e.g. 'x+'
        r_i :           str
            primary beam direction e.g. 'x+'
        detectorAxis :  list or tuple
            detector circles; e.g. ['z+', 'y-'] would mean a detector arm with
            a two rotations
        *args :         array-like
            sample, detector angles and channel numbers;
            *sAngle* sample rocking angle as numpy array, lists or Scalars.
            *dAngles* as numpy array, lists or Scalars. In total
            len(detectorAxis) values must be given starting with the outer most
            circle. All arguments must have the same shape or length.
            *channel numbers* `n1` and `n2` where the primary beam hits the
            detector with the same length as the detector values

        delta :     list of array-like, optional
            giving delta angles to correct the given ones for misalignment.
            delta must be an numpy array or list of len(*dAngles). used angles
            are than *args - delta
        UB :        array-like, optional
            orientation matrix of the sample
        wl :        float or str, optional
            x-ray wavelength in angstrom
        deg :       bool, optional
            flag to tell if angles are passed as degree (default: True)

        Returns
        -------
        ndarray
            reciprocal space position of detector pixels n1, n2 in a
            numpy.ndarray of shape (len(args) , 3)
        """

        # check detector circle argument
        if isinstance(detectorAxis, str):
            dAxis = list([detectorAxis])
        else:
            dAxis = list(detectorAxis)

        _sampleAxis_str = dAxis[-1]
        # add detector rotation around primary beam
        dAxis += [r_i]
        _detectorAxis_str = ''
        for circ in dAxis:
            _detectorAxis_str += circ

        Nd = len(dAxis)
        Nargs = Nd + 2 - 1 + 1

        # check detectorDir
        _area_detdir1 = detectorDir1
        _area_detdir2 = detectorDir2

        # parse parameter arguments
        _area_cch1 = float(params[0])
        _area_cch2 = float(params[1])
        _area_pwidth1 = float(params[2])
        _area_pwidth2 = float(params[3])
        _area_distance = float(params[4])
        _area_tiltazimuth = math.radians(params[5])
        _area_tilt = math.radians(params[6])
        _area_rot = float(params[7])
        _area_ri = xumath.getVector(r_i) * _area_distance

        # kwargs
        wl = utilities.wavelength(kwargs.get('wl', 1.))
        deg = kwargs.get('deg', True)
        UB = kwargs.get('UB', numpy.identity(3))

        delta = numpy.asarray(kwargs.get('delta', numpy.zeros(Nd)),
                              dtype=numpy.double)
        if delta.size != Nd - 1 + 1:
            raise InputError("QConversionPixel: keyword argument delta "
                             "does not have an appropriate shape")

        # prepare angular arrays from *args
        # need one sample angle and one detector angle array
        if len(args) != Nargs:
            raise InputError("QConversionPixel: wrong amount (%d) of "
                             "arguments given, number of arguments should be "
                             "%d" % (len(args), Nargs))

        try:
            Npoints = len(args[0])
        except (TypeError, IndexError):
            Npoints = 1

        sAngles = numpy.array((), dtype=numpy.double)
        for i in range(1):
            arg = args[i]
            if isinstance(arg, numbers.Number):
                arg = numpy.array([arg], dtype=numpy.double)
            elif isinstance(arg, list):
                arg = numpy.array(arg, dtype=numpy.double)
            arg = arg - delta[i]
            sAngles = numpy.concatenate((sAngles, arg))

        dAngles = numpy.array((), dtype=numpy.double)
        for i in range(1, Nd):
            arg = args[i]
            if isinstance(arg, numbers.Number):
                arg = numpy.array([arg], dtype=numpy.double)
            elif isinstance(arg, list):
                arg = numpy.array(arg, dtype=numpy.double)
            arg = arg - delta[i]
            dAngles = numpy.concatenate((dAngles, arg))
        # add detector rotation around primary beam
        dAngles = numpy.concatenate((
            dAngles,
            numpy.ones(arg.shape, dtype=numpy.double) * _area_rot))

        # read channel numbers
        n1 = numpy.array((), dtype=numpy.double)
        n2 = numpy.array((), dtype=numpy.double)

        arg = args[Nd]
        if isinstance(arg, numbers.Number):
            arg = numpy.array([arg], dtype=numpy.double)
        elif isinstance(arg, list):
            arg = numpy.array(arg, dtype=numpy.double)
        n1 = arg

        arg = args[Nd + 1]
        if isinstance(arg, numbers.Number):
            arg = numpy.array([arg], dtype=numpy.double)
        elif isinstance(arg, list):
            arg = numpy.array(arg, dtype=numpy.double)
        n2 = arg

        sAngles.shape = (1, Npoints)
        sAngles = sAngles.transpose()
        dAngles.shape = (Nd, Npoints)
        dAngles = dAngles.transpose()

        if deg:
            sAngles = radians(sAngles)
            dAngles = radians(dAngles)

        qpos = cxrayutilities.ang2q_conversion_area_pixel2(
            sAngles, dAngles, n1, n2, _area_ri, _sampleAxis_str,
            _detectorAxis_str, _area_cch1, _area_cch2, _area_pwidth1,
            _area_pwidth2, _area_detdir1, _area_detdir2, _area_tiltazimuth,
            _area_tilt, UB, wl, config.NTHREADS)

        return qpos[:, 0], qpos[:, 1], qpos[:, 2]

    def afunc(param, x, detectorDir1, detectorDir2, r_i, detectorAxis):
        """
        function for fitting the detector parameters
        basically this is a wrapper for the areapixel function

        parameters
        ----------
        param :         list or tuple
            fit parameters (cch1, cch2, pwidth1, pwidth2, distance,
            tiltazimuth, tilt, detrot, outerangle_offset, sampletilt,
            sampletiltazimuth, wavelength)
        x :             array-like
            independent variables; contains (sang, angle1, angle2, n1, n2,
            hkls) with shape (8, Npoints)
        detectorDir1 :  str
            direction of the detector (along the pixel direction 1); e.g. 'z+'
            means higher pixel numbers at larger z positions
        detectorDir2 :  str
            direction of the detector (along the pixel direction 2); e.g. 'x+'
        r_i :           str
            primary beam direction e.g. 'x+'
        detectorAxis :  list or tuple
            detector circles; e.g. ['z+', 'y-'] would mean a detector arm with
            a two rotations

        Returns
        -------
        ndarray
            reciprocal space position of detector pixels n1, n2 in a
            numpy.ndarray of shape (3, x.shape[1])
        """

        sang = x[0, :]
        angle1 = x[1, :]
        angle2 = x[2, :]
        n1 = x[3, :]
        n2 = x[4, :]
        H = x[5, :]
        K = x[6, :]
        L = x[7, :]

        # use only positive tilt and sample tilt
        param[6] = abs(param[6])
        param[9] = abs(param[9])
        wl = param[11]
        cp = math.cos(math.radians(param[10]))
        sp = math.sin(math.radians(param[10]))
        cc = math.cos(math.radians(param[9]))
        sc = math.sin(math.radians(param[9]))

        # UB matrix due to tilt at symmetric peak
        U1 = numpy.array(((cp * cc, -sp, cp * sc),
                          (sp * cc, cp, sp * sc),
                          (-sc, 0, cc)), dtype=numpy.double)
        U2 = experiment._transform.matrix
        U = numpy.dot(U1, U2)
        B = material.B
        ubmat = numpy.dot(U, B)

        (qx, qy, qz) = areapixel2(
            param[:-4], detectorDir1, detectorDir2, r_i, detectorAxis,
            sang, angle1, angle2, n1, n2, delta=[0, param[8], 0.],
            distance=1., UB=ubmat, wl=wl)

        return (qx - H) ** 2 + (qy - K) ** 2 + (qz - L) ** 2

    Npoints = len(ang1)

    # guess initial parameters
    n10 = numpy.zeros(0, dtype=numpy.double)
    n20 = n10
    ang10 = n10
    ang20 = n10
    n1s = numpy.zeros(0, dtype=numpy.double)
    n2s = n1s
    ang1s = n1s
    ang2s = n1s
    sangs = n1s

    for i in range(Npoints):
        if numpy.all(hkls[i] == (0, 0, 0)):
            n10 = numpy.append(n10, n1[i])
            n20 = numpy.append(n20, n2[i])
            ang10 = numpy.append(ang10, ang1[i])
            ang20 = numpy.append(ang20, ang2[i])
        else:
            n1s = numpy.append(n1s, n1[i])
            n2s = numpy.append(n2s, n2[i])
            ang1s = numpy.append(ang1s, ang1[i])
            ang2s = numpy.append(ang2s, ang2[i])
            sangs = numpy.append(sangs, sang[i])

    # center channel and detector pixel direction and pixel size
    (s1, i1), r1 = xumath.linregress(ang10 - start[3], n10)
    (s2, i2), r2 = xumath.linregress(ang10 - start[3], n20)
    (s3, i3), r3 = xumath.linregress(ang20, n10)
    (s4, i4), r4 = xumath.linregress(ang20, n20)

    distance = start[2]
    if ((r1 + r2 > r3 + r4) and r1 > r2) or ((r1 + r2 < r3 + r4) and r3 < r4):
        cch1 = i1
        cch2 = i4
        pwidth1 = 2 * distance / numpy.abs(s1) * math.tan(math.radians(0.5))
        pwidth2 = 2 * distance / numpy.abs(s4) * math.tan(math.radians(0.5))
    else:
        cch1 = i3
        cch2 = i2
        pwidth1 = 2 * distance / numpy.abs(s3) * math.tan(math.radians(0.5))
        pwidth2 = 2 * distance / numpy.abs(s2) * math.tan(math.radians(0.5))
    if numpy.isscalar(start[0]):
        pwidth1 = start[0]
    if numpy.isscalar(start[1]):
        pwidth2 = start[1]
    if numpy.isscalar(start[0]) or numpy.isscalar(start[1]):
        # find biggest correlation and recalculate distance
        idxmax = numpy.argmax((r1, r2, r3, r4))
        s = (s1, s2, s3, s4)[idxmax]
        distance = abs(s) / math.tan(math.radians(0.5)) / 2
        distance *= pwidth1 if idxmax < 2 else pwidth2
    tilt = abs(start[4])
    tiltazimuth = start[3]
    detrot = start[5]
    outerangle_offset = start[6]
    sampletilt = start[7]
    stazimuth = start[8]
    wavelength = start[9]
    # parameters for the fitting
    param = (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt, detrot,
             outerangle_offset, sampletilt, stazimuth, wavelength)

    # determine better start values for sample tilt and azimuth
    (qx, qy, qz) = areapixel2(
        param[:-4], detdir1, detdir2, r_i, detaxis, sangs, ang1s, ang2s,
        n1s, n2s, delta=[0, param[8], 0.], wl=wavelength)
    if debug:
        print("average qx: %.3f(%.3f)" % (numpy.average(qx), numpy.std(qx)))
        print("average qy: %.3f(%.3f)" % (numpy.average(qy), numpy.std(qy)))
        print("average qz: %.3f(%.3f)" % (numpy.average(qz), numpy.std(qz)))

    qvecav = (numpy.average(qx), numpy.average(qy), numpy.average(qz))
    sampletilt = xumath.VecAngle(experiment.Transform(experiment.ndir),
                                 qvecav, deg=True)
    stazimuth = -xumath.VecAngle(
        experiment.Transform(experiment.idir),
        qvecav - xumath.VecDot(experiment.Transform(experiment.ndir), qvecav) *
        experiment.Transform(experiment.ndir), deg=True)
    param = (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt, detrot,
             outerangle_offset, sampletilt, stazimuth, wavelength)

    if debug:
        print("initial parameters: ")
        print("primary beam / detector pixel directions / distance: %s / "
              "%s %s / %e" % (r_i, detdir1, detdir2, distance))
        print("param: (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, "
              "tilt, detrot, outerangle_offset, sampletilt, stazimuth, "
              "wavelength)")
        print("param: %.2f %.2f %10.4e %10.4e %.3f %.1f %.2f %.3f %.3f %.3f "
              "%.2f %.4f" % param)

    # set data
    x = numpy.empty((8, Npoints), dtype=numpy.double)
    x[0, :] = sang
    x[1, :] = ang1
    x[2, :] = ang2
    x[3, :] = n1
    x[4, :] = n2
    x[5, :] = hkls[:, 0]
    x[6, :] = hkls[:, 1]
    x[7, :] = hkls[:, 2]

    data = odr.Data(x, y=1)
    # define model for fitting
    model = odr.Model(afunc, extra_args=(detdir1, detdir2, r_i, detaxis),
                      implicit=True)
    # check if parameters need to be fixed
    ifixb = ()
    for i in range(len(fix)):
        ifixb += (int(not fix[i]),)

    my_odr = odr.ODR(data, model, beta0=param, ifixb=(1, 1) + ifixb,
                     ifixx=(0, 0, 0, 0, 0, 0, 0, 0),
                     stpb=(0.4, 0.4, pwidth1/50., pwidth2/50., distance/1000,
                           2, 0.125, 0.01, 0.01, 0.01, 1., 0.0001),
                     sclb=(1/abs(cch1), 1/abs(cch2), 1/pwidth1,
                           1/pwidth2, 1/distance, 1/90., 1/0.2, 1/0.2, 1/0.2,
                           1/0.1, 1/90., 1.),
                     maxit=200, ndigit=12, sstol=1e-11, partol=1e-11)
    if debug:
        my_odr.set_iprint(final=1)
        my_odr.set_iprint(iter=2)

    fit = my_odr.run()

    (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt, detrot,
     outerangle_offset, sampletilt, stazimuth, wavelength) = fit.beta
    # fix things in parameters
    tiltazimuth = tiltazimuth % 360.
    stazimuth = stazimuth % 360.
    tilt = abs(tilt)
    sampletilt = abs(sampletilt)

    final_q = afunc([cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, tilt,
                     detrot, outerangle_offset, sampletilt, stazimuth,
                     wavelength],
                    x, detdir1, detdir2, r_i, detaxis)
    final_error = numpy.mean(final_q)

    if debug:
        print("fitted parameters: (%e, %d, %s) " % (final_error, fit.info,
                                                    repr(fit.stopreason)))
        print("primary beam / detector pixel directions / distance: %s / %s "
              "%s / %e" % (r_i, detdir1, detdir2, distance))
        print("param: (cch1, cch2, pwidth1, pwidth2, distance, tiltazimuth, "
              "tilt, detrot, outerangle_offset, sampletilt, sampletiltazimuth,"
              " wavelength)")
        print("param: %.2f %.2f %10.4e %10.4e %.3f %.1f %.2f %.3f %.3f %.3f "
              "%.2f %.4f" % (cch1, cch2, pwidth1, pwidth2, distance,
                             tiltazimuth, tilt, detrot, outerangle_offset,
                             sampletilt, stazimuth, wavelength))

    if full_output:
        return final_error, (cch1, cch2, pwidth1, pwidth2, distance,
                             tiltazimuth, tilt, detrot, outerangle_offset,
                             sampletilt, stazimuth, wavelength), fit
    else:
        return final_error


#################################################
def psd_refl_align(primarybeam, angles, channels, plot=True):
    """
    function which calculates the angle at which the sample
    is parallel to the beam from various angles and detector channels
    from the reflected beam. The function can be used during the half
    beam alignment with a linear detector.

    Parameters
    ----------
    primarybeam :   int
        primary beam channel number
    angles :        list or array-like
        incidence angles
    channels :      list or array-like
        corresponding detector channels
    plot :          bool, optional
        flag to specify if a visualization of the fit is wanted default : True

    Returns
    -------
    float
        angle at which the sample is parallel to the beam

    Examples
    --------
    >>> psd_refl_align(500, [0, 0.1, 0.2, 0.3], [550, 600, 640, 700])
    """
    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis.psd_refl_'
                                                       'align')

    p, rsq = xumath.linregress(numpy.asarray(channels), numpy.asarray(angles))
    zeropos = numpy.polyval(p, primarybeam)

    if plot:
        xmin = min(min(channels), primarybeam)
        xmax = max(max(channels), primarybeam)
        ymin = min(min(angles), zeropos)
        ymax = max(max(angles), zeropos)
        # open new figure for the plot
        plt.figure()
        plt.plot(channels, angles, 'xk', ms=8., mew=2.)
        plt.plot([xmin - (xmax - xmin) * 0.1, xmax + (xmax - xmin) * 0.1],
                 numpy.polyval(p,
                               [xmin - (xmax - xmin) * 0.1,
                                xmax + (xmax - xmin) * 0.1]),
                 '-g',
                 linewidth=1.5)
        ax = plt.gca()
        plt.grid()
        ax.set_xlim(xmin - (xmax - xmin) * 0.15, xmax + (xmax - xmin) * 0.15)
        ax.set_ylim(ymin - (ymax - ymin) * 0.15, ymax + (ymax - ymin) * 0.15)
        plt.vlines(primarybeam,
                   ymin - (ymax - ymin) * 0.1,
                   ymax + (ymax - ymin) * 0.1,
                   linewidth=1.5)
        plt.xlabel("PSD Channel")
        plt.ylabel("sample angle")
        plt.tight_layout()

    if config.VERBOSITY >= config.INFO_LOW:
        print("XU.analysis.psd_refl_align: sample is parallel to beam at "
              "goniometer angle %8.4f (R^2=%6.4f)" % (zeropos, rsq))
    return zeropos


#################################################
#  miscut calculation from alignment in 2 and
#  more azimuths
#################################################
def miscut_calc(phi, aomega, zeros=None, omega0=None, plot=True):
    """
    function to calculate the miscut direction and miscut angle of a sample
    by fitting a sinusoidal function to the variation of the aligned
    omega values of more than two reflections.
    The function can also be used to fit reflectivity alignment values
    in various azimuths.

    Parameters
    ----------
    phi :       list, tuple or array-like
        azimuths in which the reflection was aligned (deg)
    aomega :    list, tuple or array-like
        aligned omega values (deg)
    zeros :     list, tuple or array-like, optional
        angles at which surface is parallel to the beam (deg). For the analysis
        the angles (aomega - zeros) are used.
    omega0 :    float, optional
        if specified the nominal value of the reflection is not included as fit
        parameter, but is fixed to the specified value. This value is MANDATORY
        if ONLY TWO AZIMUTHs are given.
    plot :      bool, optional
        flag to specify if a visualization of the fit is wanted.
        default: True

    Returns
    -------
    omega0 :    float
        the omega value of the reflection should be close to the nominal one
    phi0 :      float
        the azimuth in which the primary beam looks upstairs
    miscut :    float
        amplitude of the sinusoidal variation == miscut angle
    """
    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis.miscut_'
                                                       'calc')

    if zeros is not None:
        om = (numpy.array(aomega) - numpy.array(zeros))
    else:
        om = numpy.array(aomega)

    a = numpy.array(phi)

    if omega0 is None:
        # first guess for the parameters
        # omega0, phi0, miscut
        p0 = (om.mean(), a[om.argmax()], om.max() - om.min())

        def fitfunc(p, phi):
            return abs(p[2]) * \
                cos(radians(phi - (p[1] % 360.))) + p[0]

    else:
        # first guess for the parameters
        p0 = (a[om.argmax()], om.max() - om.min())  # omega0, phi0, miscut

        def fitfunc(p, phi):
            return abs(p[1]) * \
                cos(radians(phi - (p[0] % 360.))) + omega0

    def errfunc(p, phi, om):
        return fitfunc(p, phi) - om

    p1, success = optimize.leastsq(errfunc, p0, args=(a, om), maxfev=10000)
    if config.VERBOSITY >= config.INFO_ALL:
        print("xu.analysis.miscut_calc: leastsq optimization return value: "
              "%d" % success)

    if plot:
        plt.figure()
        plt.plot(a, om, 'xk', mew=2, ms=8)
        linx = numpy.linspace(a.min() - 45, a.min() + 360 - 45, num=1000)
        plt.plot(linx, fitfunc(p1, linx), '-g', linewidth=1.5)
        plt.grid()
        plt.xlabel("azimuth")
        plt.ylabel("aligned sample angle")

    if omega0 is None:
        ret = [p1[0], p1[1] % 360., abs(p1[2])]
    else:
        ret = [omega0] + [p1[0] % 360., abs(p1[1])]

    if config.VERBOSITY >= config.INFO_LOW:
        print("xu.analysis.miscut_calc: \n"
              "\t fitted reflection angle: %8.3f \n"
              "\t looking upstairs at phi: %8.2f \n"
              "\t miscut angle: %8.3f \n" % (ret[0], ret[1], ret[2]))

    return ret


#################################################
#  correct substrate Bragg peak position in
#  reciprocal space maps
#################################################
def fit_bragg_peak(om, tt, psd, omalign, ttalign, exphxrd, frange=(0.03, 0.03),
                   peaktype='Gauss', plot=True):
    r"""
    helper function to determine the Bragg peak position in a reciprocal
    space map used to obtain the position needed for correction of the data.
    the determination is done by fitting a two dimensional Gaussian
    (xrayutilities.math.Gauss2d) or Lorentzian
    (xrayutilities.math.Lorentz2d)

    PLEASE ALWAYS CHECK THE RESULT CAREFULLY!

    Parameters
    ----------
    om, tt :    array-like
        angular coordinates of the measurement either with size of psd or of
        psd.shape[0]
    psd :       array-like
        intensity values needed for fitting
    omalign :   float
        aligned omega value, used as first guess in the fit
    ttalign :   float
        aligned two theta values used as first guess in the fit these values
        are also used to set the range for the fit: the peak should be within
        +/-frange\AA^{-1} of those values
    exphxrd :   Experiment
        experiment class used for the conversion between angular and reciprocal
        space.
    frange :    tuple of float, optional
        data range used for the fit in both directions (see above for details
        default:(0.03, 0.03) unit: \AA^{-1})
    peaktype :  {'Gauss', 'Lorentz'}
        peak type to fit
    plot :      bool, optional
        if True (default) function will plot the result of the fit in
        comparison with the measurement.

    Returns
    -------
    omfit, ttfit :  float
        fitted angular values
    params :        list
        fit parameters (of the Gaussian/Lorentzian)
    covariance :    ndarray
        covariance matrix of the fit parameters
    """
    if plot:
        plot, plt = utilities.import_matplotlib_pyplot('XU.analysis.fit_bragg'
                                                       '_peak')

    if peaktype == 'Gauss':
        func = xumath.Gauss2d
    elif peaktype == 'Lorentz':
        func = xumath.Lorentz2d
    else:
        raise InputError("peaktype must be either 'Gauss' or 'Lorentz'")

    if om.size != psd.size:
        [qx, qy, qz] = exphxrd.Ang2Q.linear(om, tt)
    else:
        [qx, qy, qz] = exphxrd.Ang2Q(om, tt)
    [qxsub, qysub, qzsub] = exphxrd.Ang2Q(omalign, ttalign)
    params = [qysub, qzsub, 0.001, 0.001, psd.max(), 0, 0.]
    drange = [qysub - frange[0], qysub + frange[0], qzsub - frange[1],
              qzsub + frange[1]]
    params, covariance = xumath.fit_peak2d(
        qy.flatten(), qz.flatten(), psd.flatten(), params, drange,
        func, maxfev=10000)
    # correct params
    params[6] = params[6] % (numpy.pi)
    if params[5] < 0:
        params[5] = 0

    [omfit, _, _, ttfit] = exphxrd.Q2Ang((0, params[0], params[1]),
                                         trans=False, geometry="real")
    if config.VERBOSITY >= config.INFO_LOW:
        print("XU.analysis.fit_bragg_peak:fitted peak angles: \n\tom =%8.4f\n"
              "\ttt =%8.4f" % (omfit, ttfit))

    if plot:
        plt.figure()
        plt.clf()
        from ..gridder2d import Gridder2D
        gridder = Gridder2D(50, 50)
        mask = (qy.flatten() > drange[0]) * (qy.flatten() < drange[1]) * \
               (qz.flatten() > drange[2]) * (qz.flatten() < drange[3])
        gridder(qy.flatten()[mask], qz.flatten()[mask], psd.flatten()[mask])
        # calculate intensity which should be plotted
        INT = utilities.maplog(gridder.data.transpose(), 4, 0)
        QXm = gridder.xmatrix
        QZm = gridder.ymatrix
        cl = plt.contour(gridder.xaxis, gridder.yaxis,
                         utilities.maplog(func(QXm, QZm, *params), 4, 0).T,
                         8, colors='k', linestyles='solid')
        cf = plt.contourf(gridder.xaxis, gridder.yaxis, INT, 35)
        cf.collections[0].set_label('data')
        cl.collections[0].set_label('fit')
        # plt.legend() # for some reason not working?
        plt.colorbar(extend='min')
        plt.title("plot shows only coarse data! fit used raw data!")

    return omfit, ttfit, params, covariance



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

import warnings

import numpy

from .. import config, utilities
from ..exception import InputError
from . import models


class FitModel(object):
    """
    Wrapper for the lmfit Model class working for instances of LayerModel

    Typically this means that after initialization of `FitModel` you want to
    use make_params to get a `lmfit.Parameters` list which one customizes for
    fitting.

    Later on you can call `fit` and `eval` methods with those parameter list.
    """
    def __init__(self, lmodel, verbose=False, plot=False, elog=True, **kwargs):
        """
        initialization of a FitModel which uses lmfit for the actual fitting,
        and generates an according lmfit.Model internally for the given
        pre-configured LayerModel, or subclasses thereof which includes models
        for reflectivity, kinematic and dynamic diffraction.

        Parameters
        ----------
        lmodel :    LayerModel
            pre-configured instance of LayerModel or any subclass
        verbose :   bool, optional
            flag to enable verbose printing during fitting
        plot :      bool or str, optional
            flag to decide wheter an plot should be created showing the fit's
            progress. If plot is a string it will be used as figure name, which
            makes reusing the figures easier.
        elog :      bool, optional
            flag to enable a logarithmic error metric between model and data.
            Since the dynamic range of data is often large its often benefitial
            to keep this enabled.
        kwargs :    dict, optional
            additional keyword arguments are forwarded to the `simulate` method
            of `lmodel`
        """
        self.verbose = verbose
        self.plot = plot
        self.elog = elog
        lmfit = utilities.import_lmfit('XU.simpack')

        assert isinstance(lmodel, models.LayerModel)
        self.lmodel = lmodel
        # generate dynamic function for model evalution
        funcstr = "def func(x, "
        # add LayerModel parameters
        for p in self.lmodel.fit_paramnames:
            funcstr += "{}, ".format(p)
        # add LayerStack parameters
        for layer in self.lmodel.lstack:
            for param in self.lmodel.lstack_params:
                funcstr += '{}_{}, '.format(layer.name, param)
            if self.lmodel.lstack_structural_params:
                for param in layer._structural_params:
                    funcstr += '{}_{}, '.format(layer.name, param)
        funcstr += "lmodel=self.lmodel, **kwargs):\n"
        # define modelfunc content
        for p in self.lmodel.fit_paramnames:
            funcstr += "    setattr(lmodel, '{}', {})\n".format(p, p)
        for i, l in enumerate(self.lmodel.lstack):
            for param in self.lmodel.lstack_params:
                varname = '{}_{}'.format(l.name, param)
                cmd = "    setattr(lmodel.lstack[{}], '{}', {})\n"
                funcstr += cmd.format(i, param, varname)
            if self.lmodel.lstack_structural_params:
                for param in l._structural_params:
                    varname = '{}_{}'.format(l.name, param)
                    cmd = "    setattr(lmodel.lstack[{}], '{}', {})\n"
                    funcstr += cmd.format(i, param, varname)
        # perform actual model calculation
        funcstr += "    return lmodel.simulate(x, **kwargs)"

        namespace = {'self': self}
        exec(funcstr, {'lmodel': self.lmodel}, namespace)
        self.func = namespace['func']
        self.emetricfunc = numpy.log10 if self.elog else lambda x: x

        def _residual(params, data, weights, **kwargs):
            """
            Return the residual. This is a (simplified, only real values)
            reimplementation of the lmfit.Model._residual function which adds
            the possibility of a logarithmic error metric.

            Default residual: (data-model)*weights.
            """
            scale = self.emetricfunc
            model = scale(self.eval(params, **kwargs))
            sdata = scale(data)
            mask = numpy.logical_and(numpy.isfinite(model),
                                     numpy.isfinite(sdata))
            diff = model[mask] - sdata[mask]
            if weights is not None and scale(1) == 1:
                diff *= weights
            return numpy.asarray(diff).ravel()

        self.lmm = lmfit.Model(self.func,
                               name=self.lmodel.__class__.__name__, **kwargs)
        self.lmm._residual = _residual
        for method in ('set_param_hint', 'print_param_hints', 'eval',
                       'make_params'):
            setattr(self, method, getattr(self.lmm, method))
        # set default parameter hints
        self._default_hints()
        self.set_fit_limits()

    def set_fit_limits(self, xmin=-numpy.inf, xmax=numpy.inf, mask=None):
        """
        set fit limits. If mask is given it must have the same size as the
        `data` and `x` variables given to fit. If mask is None it will be
        generated from xmin and xmax.

        Parameters
        ----------
        xmin :  float, optional
            minimum value of x-values to include in the fit
        xmax :  float, optional
            maximum value of x-values to include in the fit
        mask :  boolean array, optional
            mask to be used for the data given to the fit
        """
        self.mask = mask
        self.xmin = xmin
        self.xmax = xmax

    def fit(self, data, params, x, weights=None, fit_kws=None, **kwargs):
        """
        wrapper around lmfit.Model.fit which enables plotting during the
        fitting

        Parameters
        ----------
        data :      ndarray
            experimental values
        params :    lmfit.Parameters
            list of parameters for the fit, use make_params for generation
        x :         ndarray
            independent variable (incidence angle or q-position depending on
            the model)
        weights :   ndarray, optional
            values of weights for the fit, same size as data
        fit_kws :   dict, optional
            Options to pass to the minimizer being used
        kwargs :    dict, optional
            keyword arguments which are passed to lmfit.Model.fit

        Returns
        -------
        lmfit.ModelResult
        """
        class FitPlot(object):
            def __init__(self, figname, logscale):
                self.figname = figname
                self.logscale = logscale
                if not self.figname:
                    self.plot = False
                else:
                    f, plt = utilities.import_matplotlib_pyplot('XU.simpack')
                    self.plt = plt
                    self.plot = f

            def plot_init(self, x, data, weights, model, mask, verbose):
                if not self.plot:
                    return
                self.plt.ion()
                if isinstance(self.figname, str):
                    self.fig = self.plt.figure(self.figname)
                else:
                    self.fig = self.plt.figure('XU:FitModel')
                self.plt.clf()
                self.ax = self.plt.subplot(111)

                if weights is not None:
                    eline = self.ax.errorbar(
                        x, data, yerr=1/weights, ecolor='0.3', fmt='ok',
                        errorevery=int(x.size/80), label='data')[0]
                else:
                    eline, = self.ax.plot(x, data, 'ok', label='data')
                if verbose:
                    init, = self.ax.plot(
                        x, model, '-', color='0.5', label='initial')
                if eline:
                    self.zord = eline.zorder+2
                else:
                    self.zord = 1
                if self.logscale:
                    self.ax.set_yscale("log")
                self.fline = None

            def showplot(self, xlab=self.lmodel.xlabelstr,
                         ylab='Intensity (arb. u.)'):
                if not self.plot:
                    return
                self.plt.xlabel(xlab)
                self.plt.ylabel(ylab)
                self.plt.legend()
                self.fig.set_tight_layout(True)
                self.plt.show()

            def updatemodelline(self, x, newmodel):
                if not self.plot:
                    return
                try:
                    self.plt.sca(self.ax)
                except ValueError:
                    return
                if self.fline is None:
                    self.fline, = self.ax.plot(
                        x, newmodel, '-r', lw=2, label='fit', zorder=self.zord)
                else:
                    self.fline.set_data(x, newmodel)
                canvas = self.fig.canvas  # see plt.draw function (avoid show!)
                canvas.draw_idle()
                canvas.start_event_loop(0.001)

            def addfullmodelline(self, x, y):
                if not self.plot:
                    return
                self.ax.plot(x, y, '-g', lw=1, label='full model',
                             zorder=self.zord-1)

        if self.mask:
            mask = self.mask
        else:
            mask = numpy.logical_and(x >= self.xmin, x <= self.xmax)
        mweights = weights
        if mweights is not None:
            mweights = weights[mask]

        # create initial plot
        self.fitplot = FitPlot(self.plot, self.elog)
        initmodel = self.eval(params, x=x, **kwargs)
        self.fitplot.plot_init(x, data, weights, initmodel, mask, self.verbose)
        self.fitplot.showplot()

        # create callback function
        def cb_func(params, niter, resid, *args, **kwargs):
            if self.verbose:
                print('{:04d} {:12.3e}'.format(niter, numpy.sum(resid**2)))
            if self.fitplot.plot and niter % 20 == 0:
                self.fitplot.updatemodelline(kwargs['x'],
                                             self.eval(params, **kwargs))

        # perform fitting
        res = self.lmm.fit(data[mask], params, x=x[mask], weights=mweights,
                           fit_kws=fit_kws, iter_cb=cb_func, **kwargs)

        # final update of plot
        if self.fitplot.plot:
            try:
                self.fitplot.plt.sca(self.fitplot.ax)
            except ValueError:
                self.fitplot.plot_init(x, data, weights, initmodel, mask,
                                       self.verbose)
            fittedmodel = self.eval(res.params, x=x, **kwargs)
            self.fitplot.addfullmodelline(x, fittedmodel)
            self.fitplot.updatemodelline(x[mask], fittedmodel[mask])
            self.fitplot.showplot()

        return res

    def _default_hints(self):
        """
        set useful hints for parameters all LayerModels have
        """
        # general parameters
        for pn in self.lmodel.fit_paramnames:
            self.set_param_hint(pn, value=getattr(self.lmodel, pn), vary=False)
        for pn in ('I0', 'background'):
            self.set_param_hint(pn, vary=True, min=0)
        self.set_param_hint('resolution_width', min=0, vary=False)
        self.set_param_hint('energy', min=1000, vary=False)

        # parameters of the layerstack
        for lay in self.lmodel.lstack:
            for param in self.lmodel.lstack_params:
                varname = '{}_{}'.format(lay.name, param)
                self.set_param_hint(varname, value=getattr(lay, param), min=0)
                if param == 'density':
                    self.set_param_hint(varname, max=1.5*lay.material.density)
                if param == 'thickness':
                    self.set_param_hint(varname, max=2*lay.thickness)
                if param == 'roughness':
                    self.set_param_hint(varname, max=50)
            if self.lmodel.lstack_structural_params:
                for param in lay._structural_params:
                    varname = '{}_{}'.format(lay.name, param)
                    self.set_param_hint(varname, value=getattr(lay, param),
                                        vary=False)
                    if 'occupation' in param:
                        self.set_param_hint(varname, min=0, max=1)
                    if 'biso' in param:
                        self.set_param_hint(varname, min=0, max=5)
        if self.lmodel.lstack[0].thickness == numpy.inf:
            varname = '{}_{}'.format(self.lmodel.lstack[0].name, 'thickness')
            self.set_param_hint(varname, vary=False)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009-2010, 2013
#               Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2009 Mario Keplinger <mario.keplinger@jku.at>
# Copyright (C) 2009-2021 Dominik Kriegner <dominik.kriegner@gmail.com>

import abc

import numpy

from . import config, cxrayutilities, utilities
from .exception import InputError


def delta(min_value, max_value, n):
    """
    Compute the stepsize along an axis of a grid.

    Parameters
    ----------
   min_value :	 axis minimum value
   max_value :	 axis maximum value
   n :	 number of steps
    """
    if n != 1:
        return (float(max_value) - float(min_value)) / float(n - 1)
    else:
        return numpy.inf


def axis(min_value, max_value, n):
    """
    Compute the a grid axis.

    Parameters
    ----------
    min_value : float
        axis minimum value
    max_value :	float
        axis maximum value
    n :	        int
        number of steps
    """

    if n != 1:
        d = delta(min_value, max_value, n)
        a = min_value + d * numpy.arange(0, n)
    else:
        a = (min_value + max_value) / 2.

    return a


def ones(*args):
    """
    Compute ones for matrix generation. The shape is determined by the number
    of input arguments.
    """
    return numpy.ones(args, dtype=numpy.double)


class Gridder(utilities.ABC):
    """
    Basis class for gridders in xrayutilities. A gridder is a function mapping
    irregular spaced data onto a regular grid by binning the data into equally
    sized elements.

    There are different ways of defining the regular grid of a Gridder. In
    xrayutilities the data bins extend beyond the data range in the input data,
    but the given position being the center of these bins, extends from the
    minimum to the maximum of the data!  The main motivation for this was to
    create a Gridder, which when feeded with N equidistant data points and
    gridded with N bins would not change the data position (not the case with
    numpy.histogramm functions!). Of course this leads to the fact that for
    homogeneous point density the first and last bin in any direction are not
    filled as the other bins.

    A different definition is used by numpy histogram functions where the bins
    extend only to the end of the data range. (see numpy histogram,
    histrogram2d, ...)
    """

    def __init__(self):
        """
        Constructor defining default properties of any Gridder class
        """

        self.flags = 0
        # by default every call to gridder will start a new gridding
        self.keep_data = False
        self.normalize = True
        # flag to allow for sequential gridding with fixed data range
        self.fixed_range = False

        # no data initialization necessary in c-code
        self.flags = utilities.set_bit(self.flags, 0)

        if not hasattr(self, '_gdata'):
            self._gdata = numpy.empty(0)
        if not hasattr(self, '_gnorm'):
            self._gnorm = numpy.empty(0)

        if config.VERBOSITY >= config.INFO_ALL:
            self.flags = utilities.set_bit(self.flags, 3)

    @abc.abstractmethod
    def __call__(self):
        """
        abstract call method which every implementation of a Gridder has to
        override
        """
        pass

    def Normalize(self, bool):
        """
        set or unset the normalization flag.  Normalization needs to be done to
        obtain proper gridding but may want to be disabled in certain cases
        when sequential gridding is performed
        """
        if bool not in [False, True]:
            raise TypeError("Normalize flag must be a boolan value "
                            "(True/False)!")
        self.normalize = bool
        if bool:
            self.flags = utilities.clear_bit(self.flags, 2)
        else:
            self.flags = utilities.set_bit(self.flags, 2)

    def KeepData(self, bool):
        if bool not in [False, True]:
            raise TypeError("Keep Data flag must be a boolan value"
                            "(True/False)!")

        self.keep_data = bool

    def __get_data(self):
        """
        return gridded data (performs normalization if switched on)
        """
        if self.normalize:
            tmp = numpy.copy(self._gdata)
            mask = (self._gnorm != 0)
            tmp[mask] /= self._gnorm[mask].astype(numpy.double)
            return tmp
        else:
            return self._gdata.copy()

    data = property(__get_data)

    def _prepare_array(self, a):
        """
        prepare array for passing to c-code
        """
        if isinstance(a, (list, tuple, float, int)):
            a = numpy.asarray(a)
        return a.reshape(a.size)

    def Clear(self):
        """
        Clear so far gridded data to reuse this instance of the Gridder
        """
        self._gdata[...] = 0
        self._gnorm[...] = 0


class Gridder1D(Gridder):

    def __init__(self, nx):
        Gridder.__init__(self)
        if nx <= 0:
            raise InputError('nx must be a positiv integer!')

        self.nx = nx
        self.xmin = 0
        self.xmax = 0
        self._gdata = numpy.zeros(nx, dtype=numpy.double)
        self._gnorm = numpy.zeros(nx, dtype=numpy.double)

    def savetxt(self, filename, header=''):
        """
        save gridded data to a txt file with two columns. The first column is
        the data coordinate and the second the corresponding data value

        Parameters
        ----------
        filename :  str
            output filename
        header :    str, optional
            optional header for the data file.
        """
        numpy.savetxt(filename, numpy.vstack((self.xaxis, self.data)).T,
                      header=header, fmt='%.6g %.4g')

    def __get_xaxis(self):
        """
        Returns the xaxis of the gridder
        the returned values correspond to the center of the data bins used by
        the gridding algorithm
        """
        return axis(self.xmin, self.xmax, self.nx)

    xaxis = property(__get_xaxis)

    def dataRange(self, min, max, fixed=True):
        """
        define minimum and maximum data range, usually this is deduced
        from the given data automatically, however, for sequential
        gridding it is useful to set this before the first call of the
        gridder. data outside the range are simply ignored

        Parameters
        ----------
        min :   float
            minimum value of the gridding range
        max :   float
            maximum value of the gridding range
        fixed : bool, optional
            flag to turn fixed range gridding on (True (default)) or off
            (False)
        """
        self.fixed_range = fixed
        self.xmin = min
        self.xmax = max

    def _checktransinput(self, x, data):
        """
        common checks and reshape commands for the input data. This function
        checks the data type and shape of the input data.
        """
        if not self.keep_data:
            self.Clear()

        x = self._prepare_array(x)
        data = self._prepare_array(data)

        if x.size != data.size:
            raise InputError("XU.%s: size of given datasets (x, data)"
                             " is not equal!" % self.__class__.__name__)

        if not self.fixed_range:
            # assume that with setting keep_data the user wants to call the
            # gridder more often and obtain a reasonable result
            self.dataRange(x.min(), x.max(), self.keep_data)

        return x, data

    def __call__(self, x, data):
        """
        Perform gridding on a set of data. After running the gridder
        the 'data' object in the class is holding the gridded data.

        Parameters
        ----------
        x :	 ndarray
            numpy array of arbitrary shape with x positions
        data :	 ndarray
            numpy array of arbitrary shape with data values
        """
        x, data = self._checktransinput(x, data)
        # remove normalize flag for C-code, normalization is always performed
        # in python
        flags = utilities.set_bit(self.flags, 2)
        cxrayutilities.gridder1d(x, data, self.nx, self.xmin, self.xmax,
                                 self._gdata, self._gnorm, flags)


class FuzzyGridder1D(Gridder1D):
    """
    An 1D binning class considering every data point to have a finite width.
    If necessary one data point will be split fractionally over different
    data bins. This is numerically more effort but represents better the
    typical case of a experimental data, which do not represent a mathematical
    point but have a finite width (e.g. X-ray data from a 1D detector).
    """

    def __call__(self, x, data, width=None):
        """
        Perform gridding on a set of data. After running the gridder
        the 'data' object in the class is holding the gridded data.

        Parameters
        ----------
        x :	 ndarray
            numpy array of arbitrary shape with x positions
        data :	 ndarray
            numpy array of arbitrary shape with data values
        width :	 float, optional
            width of one data point. If not given half the bin size will be
            used.
        """
        x, data = self._checktransinput(x, data)

        if not width:
            width = delta(self.xmin, self.xmax, self.nx) / 2.

        # remove normalize flag for C-code, normalization is always performed
        # in python
        flags = utilities.set_bit(self.flags, 2)
        cxrayutilities.fuzzygridder1d(x, data, self.nx, self.xmin, self.xmax,
                                      self._gdata, self._gnorm, width, flags)


class npyGridder1D(Gridder1D):

    def __get_xaxis(self):
        """
        Returns the xaxis of the gridder
        the returned values correspond to the center of the data bins used by
        the numpy.histogram function
        """
        # no -1 here to be consistent with numpy.histogram
        dx = (float(self.xmax - self.xmin)) / float(self.nx)
        ax = self.xmin + dx * numpy.arange(0, self.nx) + dx / 2.
        return ax

    xaxis = property(__get_xaxis)

    def __call__(self, x, data):
        """
        Perform gridding on a set of data. After running the gridder
        the 'data' object in the class is holding the gridded data.

        Parameters
        ----------
        x :	 ndarray
            numpy array of arbitrary shape with x positions
        data :	 ndarray
            numpy array of arbitrary shape with data values
        """

        x, data = self._checktransinput(x, data)

        # use only non-NaN data values
        mask = numpy.invert(numpy.isnan(data))
        ldata = data[mask]
        lx = x[mask]

        if not self.fixed_range:
            # assume that with setting keep_data the user wants to call the
            # gridder more often and obtain a reasonable result
            self.dataRange(lx.min(), lx.max(), self.keep_data)

        # grid the data using numpy histogram
        tmpgdata, bins = numpy.histogram(lx, weights=ldata, bins=self.nx,
                                         range=(self.xmin, self.xmax))
        tmpgnorm, bins = numpy.histogram(lx, bins=self.nx,
                                         range=(self.xmin, self.xmax))
        if self.keep_data:
            self._gnorm += tmpgnorm
            self._gdata += tmpgdata
        else:
            self._gnorm = tmpgnorm
            self._gdata = tmpgdata
